# Copyright (c), Michael DeHaan <michael.dehaan@gmail.com>, 2012-2013
# Copyright (c), Toshio Kuratomi <tkuratomi@ansible.com> 2016
# Simplified BSD License (see licenses/simplified_bsd.txt or https://opensource.org/licenses/BSD-2-Clause)

from __future__ import annotations

import copy
import json
import sys
import typing as t

# Used for determining if the system is running a new enough python version
# and should only restrict on our documented minimum versions
_PY_MIN = (3, 8)

if sys.version_info < _PY_MIN:
    print(json.dumps(dict(
        failed=True,
        msg=f"ansible-core requires a minimum of Python version {'.'.join(map(str, _PY_MIN))}. Current version: {''.join(sys.version.splitlines())}",
    )))
    sys.exit(1)

# Ansible modules can be written in any language.
# The functions available here can be used to do many common tasks,
# to simplify development of Python modules.

import __main__
import atexit
import dataclasses as _dataclasses
import errno
import grp
import fcntl
import locale
import os
import pwd
import platform
import re
import select
import selectors
import shlex
import shutil
import stat
import subprocess
import tempfile
import time
import traceback

from functools import reduce

try:
    import syslog
    HAS_SYSLOG = True
except ImportError:
    HAS_SYSLOG = False

_UNSET = t.cast(t.Any, object())

try:
    from systemd import journal, daemon as systemd_daemon
    # Makes sure that systemd.journal has method sendv()
    # Double check that journal has method sendv (some packages don't)
    # check if the system is running under systemd
    has_journal = hasattr(journal, 'sendv') and systemd_daemon.booted()
except (ImportError, AttributeError):
    # AttributeError would be caused from use of .booted() if wrong systemd
    has_journal = False

HAVE_SELINUX = False
try:
    from ansible.module_utils.compat import selinux
    HAVE_SELINUX = True
except ImportError:
    pass

# Python2 & 3 way to get NoneType
NoneType = type(None)

from ._internal import _traceback, _errors, _debugging, _deprecator, _messages

from .common.text.converters import (
    to_native,
    to_bytes,
    to_text,
    jsonify,
    container_to_bytes as json_dict_unicode_to_bytes,
    container_to_text as json_dict_bytes_to_unicode,
)

from ansible.module_utils.common.arg_spec import ModuleArgumentSpecValidator

from ansible.module_utils.common.text.formatters import (
    lenient_lowercase,
    bytes_to_human,
    human_to_bytes,
    SIZE_RANGES,
)

from ansible.module_utils.common import json as _common_json

import hashlib


def _get_available_hash_algorithms():
    """Return a dictionary of available hash function names and their associated function."""
    algorithms = {}
    for algorithm_name in hashlib.algorithms_available:
        algorithm_func = getattr(hashlib, algorithm_name, None)
        if algorithm_func:
            try:
                # Make sure the algorithm is actually available for use.
                # Not all algorithms listed as available are actually usable.
                # For example, md5 is not available in FIPS mode.
                algorithm_func()
            except Exception:
                pass
            else:
                algorithms[algorithm_name] = algorithm_func

    return algorithms


AVAILABLE_HASH_ALGORITHMS = _get_available_hash_algorithms()

from ansible.module_utils.common import json as _json

from ansible.module_utils.six.moves.collections_abc import (
    KeysView,
    Mapping, MutableMapping,
    Sequence, MutableSequence,
    Set, MutableSet,
)
from ansible.module_utils.common.locale import get_best_parsable_locale
from ansible.module_utils.common.process import get_bin_path
from ansible.module_utils.common.file import (
    _PERM_BITS as PERM_BITS,
    _DEFAULT_PERM as DEFAULT_PERM,
    is_executable,
    format_attributes,
    get_flags_from_attributes,
    FILE_ATTRIBUTES,
    S_IXANY,
    S_IRWU_RWG_RWO,
)
from ansible.module_utils.common.sys_info import (
    get_distribution,
    get_distribution_version,
    get_platform_subclass,
)
from ansible.module_utils.common.parameters import (
    env_fallback,
    remove_values,
    sanitize_keys,
    DEFAULT_TYPE_VALIDATORS,
    PASS_VARS,
    PASS_BOOLS,
)

from ansible.module_utils.errors import AnsibleFallbackNotFound, AnsibleValidationErrorMultiple, UnsupportedError
from ansible.module_utils.common.validation import (
    check_missing_parameters,
    safe_eval,
)
from ansible.module_utils.common._utils import get_all_subclasses as _get_all_subclasses
from ansible.module_utils.parsing.convert_bool import BOOLEANS, BOOLEANS_FALSE, BOOLEANS_TRUE, boolean
from ansible.module_utils.common.warnings import (
    deprecate,
    error_as_warning,
    get_deprecations,
    get_warnings,
    warn,
)

# Note: When getting Sequence from collections, it matches with strings. If
# this matters, make sure to check for strings before checking for sequencetype
SEQUENCETYPE = frozenset, KeysView, Sequence

PASSWORD_MATCH = re.compile(r'^(?:.+[-_\s])?pass(?:[-_\s]?(?:word|phrase|wrd|wd)?)(?:[-_\s].+)?$', re.I)

imap = map

# Internal global holding passed in params.  This is consulted in case
# multiple AnsibleModules are created.  Otherwise each AnsibleModule would
# attempt to read from stdin.  Other code should not use this directly as it
# is an internal implementation detail
_ANSIBLE_ARGS: bytes | None = None
_ANSIBLE_PROFILE: str | None = None
_PARSED_MODULE_ARGS: dict[str, t.Any] | None = None


FILE_COMMON_ARGUMENTS = dict(
    # These are things we want. About setting metadata (mode, ownership, permissions in general) on
    # created files (these are used by set_fs_attributes_if_different and included in
    # load_file_common_arguments)
    mode=dict(type='raw'),
    owner=dict(type='str'),
    group=dict(type='str'),
    seuser=dict(type='str'),
    serole=dict(type='str'),
    selevel=dict(type='str'),
    setype=dict(type='str'),
    attributes=dict(type='str', aliases=['attr']),
    unsafe_writes=dict(type='bool', default=False, fallback=(env_fallback, ['ANSIBLE_UNSAFE_WRITES'])),  # should be available to any module using atomic_move
)

PASSWD_ARG_RE = re.compile(r'^[-]{0,2}pass[-]?(word|wd)?')

# Used for parsing symbolic file perms
MODE_OPERATOR_RE = re.compile(r'[+=-]')
USERS_RE = re.compile(r'^[ugo]+$')
PERMS_RE = re.compile(r'^[rwxXstugo]*$')


#
# Deprecated functions
#

def get_platform():
    """
    **Deprecated** Use :py:func:`platform.system` directly.

    :returns: Name of the platform the module is running on in a native string

    Returns a native string that labels the platform ("Linux", "Solaris", etc). Currently, this is
    the result of calling :py:func:`platform.system`.
    """
    return platform.system()

# End deprecated functions


#
# Compat shims
#

def load_platform_subclass(cls, *args, **kwargs):
    """**Deprecated**: Use ansible.module_utils.common.sys_info.get_platform_subclass instead"""
    platform_cls = get_platform_subclass(cls)
    return super(cls, platform_cls).__new__(platform_cls)


def get_all_subclasses(cls):
    """**Deprecated**: Use ansible.module_utils.common._utils.get_all_subclasses instead"""
    return list(_get_all_subclasses(cls))


# End compat shims


def heuristic_log_sanitize(data, no_log_values=None):
    """ Remove strings that look like passwords from log messages """
    # Currently filters:
    # user:pass@foo/whatever and http://username:pass@wherever/foo
    # This code has false positives and consumes parts of logs that are
    # not passwds

    # begin: start of a passwd containing string
    # end: end of a passwd containing string
    # sep: char between user and passwd
    # prev_begin: where in the overall string to start a search for
    #   a passwd
    # sep_search_end: where in the string to end a search for the sep
    data = to_native(data)

    output = []
    begin = len(data)
    prev_begin = begin
    sep = 1
    while sep:
        # Find the potential end of a passwd
        try:
            end = data.rindex('@', 0, begin)
        except ValueError:
            # No passwd in the rest of the data
            output.insert(0, data[0:begin])
            break

        # Search for the beginning of a passwd
        sep = None
        sep_search_end = end
        while not sep:
            # URL-style username+password
            try:
                begin = data.rindex('://', 0, sep_search_end)
            except ValueError:
                # No url style in the data, check for ssh style in the
                # rest of the string
                begin = 0
            # Search for separator
            try:
                sep = data.index(':', begin + 3, end)
            except ValueError:
                # No separator; choices:
                if begin == 0:
                    # Searched the whole string so there's no password
                    # here.  Return the remaining data
                    output.insert(0, data[0:prev_begin])
                    break
                # Search for a different beginning of the password field.
                sep_search_end = begin
                continue
        if sep:
            # Password was found; remove it.
            output.insert(0, data[end:prev_begin])
            output.insert(0, '********')
            output.insert(0, data[begin:sep + 1])
            prev_begin = begin

    output = ''.join(output)
    if no_log_values:
        output = remove_values(output, no_log_values)
    return output


def _load_params():
    """ read the modules parameters and store them globally.

    This function may be needed for certain very dynamic custom modules which
    want to process the parameters that are being handed the module.  Since
    this is so closely tied to the implementation of modules we cannot
    guarantee API stability for it (it may change between versions) however we
    will try not to break it gratuitously.  It is certainly more future-proof
    to call this function and consume its outputs than to implement the logic
    inside it as a copy in your own code.
    """
    global _ANSIBLE_ARGS, _ANSIBLE_PROFILE

    if _ANSIBLE_ARGS is None:
        _ANSIBLE_ARGS, _ANSIBLE_PROFILE = _debugging.load_params()

    buffer = _ANSIBLE_ARGS
    profile = _ANSIBLE_PROFILE

    if not profile:
        raise Exception("No serialization profile was specified.")

    try:
        decoder = _json.get_module_decoder(profile, _json.Direction.CONTROLLER_TO_MODULE)
        params = json.loads(buffer.decode(), cls=decoder)
    except Exception as ex:
        raise Exception("Failed to decode JSON module parameters.") from ex

    if (ansible_module_args := params.get('ANSIBLE_MODULE_ARGS', _UNSET)) is _UNSET:
        raise Exception("ANSIBLE_MODULE_ARGS not provided.")

    global _PARSED_MODULE_ARGS

    _PARSED_MODULE_ARGS = copy.deepcopy(ansible_module_args)  # AnsibleModule mutates the returned dict, so a copy is needed

    return ansible_module_args


def missing_required_lib(library, reason=None, url=None):
    hostname = platform.node()
    msg = "Failed to import the required Python library (%s) on %s's Python %s." % (library, hostname, sys.executable)
    if reason:
        msg += " This is required %s." % reason
    if url:
        msg += " See %s for more info." % url

    msg += (" Please read the module documentation and install it in the appropriate location."
            " If the required library is installed, but Ansible is using the wrong Python interpreter,"
            " please consult the documentation on ansible_python_interpreter")
    return msg


class AnsibleModule(object):
    def __init__(self, argument_spec, bypass_checks=False, no_log=False,
                 mutually_exclusive=None, required_together=None,
                 required_one_of=None, add_file_common_args=False,
                 supports_check_mode=False, required_if=None, required_by=None):

        """
        Common code for quickly building an ansible module in Python
        (although you can write modules with anything that can return JSON).

        See :ref:`developing_modules_general` for a general introduction
        and :ref:`developing_program_flow_modules` for more detailed explanation.
        """

        self._name = os.path.basename(__file__)  # initialize name until we can parse from options
        self.argument_spec = argument_spec
        self.supports_check_mode = supports_check_mode
        self.check_mode = False
        self.bypass_checks = bypass_checks
        self.no_log = no_log

        self.mutually_exclusive = mutually_exclusive
        self.required_together = required_together
        self.required_one_of = required_one_of
        self.required_if = required_if
        self.required_by = required_by
        self.cleanup_files = []
        self._debug = False
        self._diff = False
        self._socket_path = None
        self._shell = None
        self._syslog_facility = 'LOG_USER'
        self._verbosity = 0
        # May be used to set modifications to the environment for any
        # run_command invocation
        self.run_command_environ_update = {}
        self._clean = {}

        self.aliases = {}
        self._legal_inputs = []
        self._options_context = list()
        self._tmpdir = None

        if add_file_common_args:
            for k, v in FILE_COMMON_ARGUMENTS.items():
                if k not in self.argument_spec:
                    self.argument_spec[k] = v

        # Save parameter values that should never be logged
        self.no_log_values = set()

        # check the locale as set by the current environment, and reset to
        # a known valid (LANG=C) if it's an invalid/unavailable locale
        self._check_locale()

        self._load_params()
        self._set_internal_properties()

        self.validator = ModuleArgumentSpecValidator(self.argument_spec,
                                                     self.mutually_exclusive,
                                                     self.required_together,
                                                     self.required_one_of,
                                                     self.required_if,
                                                     self.required_by,
                                                     )

        self.validation_result = self.validator.validate(self.params)
        self.params.update(self.validation_result.validated_parameters)
        self.no_log_values.update(self.validation_result._no_log_values)
        self.aliases.update(self.validation_result._aliases)

        try:
            error = self.validation_result.errors[0]
            if isinstance(error, UnsupportedError) and self._ignore_unknown_opts:
                error = None
        except IndexError:
            error = None

        # Fail for validation errors, even in check mode
        if error:
            msg = self.validation_result.errors.msg
            if isinstance(error, UnsupportedError):
                msg = "Unsupported parameters for ({name}) {kind}: {msg}".format(name=self._name, kind='module', msg=msg)

            self.fail_json(msg=msg)

        if self.check_mode and not self.supports_check_mode:
            self.exit_json(skipped=True, msg="remote module (%s) does not support check mode" % self._name)

        # This is for backwards compatibility only.
        self._CHECK_ARGUMENT_TYPES_DISPATCHER = DEFAULT_TYPE_VALIDATORS

        if not self.no_log:
            self._log_invocation()

        # selinux state caching
        self._selinux_enabled = None
        self._selinux_mls_enabled = None
        self._selinux_initial_context = None

        # finally, make sure we're in a logical working dir
        self._set_cwd()

    @property
    def tmpdir(self):
        # if _ansible_tmpdir was not set and we have a remote_tmp,
        # the module needs to create it and clean it up once finished.
        # otherwise we create our own module tmp dir from the system defaults
        if self._tmpdir is None:
            basedir = None

            if self._remote_tmp is not None:
                basedir = os.path.expanduser(os.path.expandvars(self._remote_tmp))

            if basedir is not None and not os.path.exists(basedir):
                try:
                    os.makedirs(basedir, mode=0o700)
                except OSError as ex:
                    self.error_as_warning(
                        msg=f"Unable to use {basedir!r} as temporary directory, falling back to system default.",
                        exception=ex,
                    )
                    basedir = None
                else:
                    self.warn("Module remote_tmp %s did not exist and was "
                              "created with a mode of 0700, this may cause"
                              " issues when running as another user. To "
                              "avoid this, create the remote_tmp dir with "
                              "the correct permissions manually" % basedir)

            basefile = "ansible-moduletmp-%s-" % time.time()
            try:
                tmpdir = tempfile.mkdtemp(prefix=basefile, dir=basedir)
            except OSError as ex:
                raise Exception(
                    f"Failed to create remote module tmp path at dir {basedir!r} "
                    f"with prefix {basefile!r}.",
                ) from ex
            if not self._keep_remote_files:
                atexit.register(shutil.rmtree, tmpdir)
            self._tmpdir = tmpdir

        return self._tmpdir

    def warn(
        self,
        warning: str,
        *,
        help_text: str | None = None,
    ) -> None:
        _skip_stackwalk = True

        warn(
            warning=warning,
            help_text=help_text,
        )

    def error_as_warning(
        self,
        msg: str | None,
        exception: BaseException,
        *,
        help_text: str | None = None,
    ) -> None:
        """Display an exception as a warning."""
        _skip_stackwalk = True

        error_as_warning(
            msg=msg,
            exception=exception,
            help_text=help_text,
        )

    def deprecate(
        self,
        msg: str,
        version: str | None = None,
        date: str | None = None,
        collection_name: str | None = None,
        *,
        deprecator: _messages.PluginInfo | None = None,
        help_text: str | None = None,
    ) -> None:
        """
        Record a deprecation warning to be returned with the module result.
        Most callers do not need to provide `collection_name` or `deprecator` -- but provide only one if needed.
        Specify `version` or `date`, but not both.
        If `date` is a string, it must be in the form `YYYY-MM-DD`.
        """
        _skip_stackwalk = True

        deprecate(  # pylint: disable=ansible-deprecated-date-not-permitted,ansible-deprecated-unnecessary-collection-name
            msg=msg,
            version=version,
            date=date,
            deprecator=_deprecator.get_best_deprecator(deprecator=deprecator, collection_name=collection_name),
            help_text=help_text,
        )

    def load_file_common_arguments(self, params, path=None):
        """
        many modules deal with files, this encapsulates common
        options that the file module accepts such that it is directly
        available to all modules and they can share code.

        Allows to overwrite the path/dest module argument by providing path.
        """

        if path is None:
            path = params.get('path', params.get('dest', None))
        if path is None:
            return {}
        else:
            path = os.path.expanduser(os.path.expandvars(path))

        b_path = to_bytes(path, errors='surrogate_or_strict')
        # if the path is a symlink, and we're following links, get
        # the target of the link instead for testing
        if params.get('follow', False) and os.path.islink(b_path):
            b_path = os.path.realpath(b_path)
            path = to_native(b_path)

        mode = params.get('mode', None)
        owner = params.get('owner', None)
        group = params.get('group', None)

        # selinux related options
        seuser = params.get('seuser', None)
        serole = params.get('serole', None)
        setype = params.get('setype', None)
        selevel = params.get('selevel', None)
        secontext = [seuser, serole, setype]

        if self.selinux_mls_enabled():
            secontext.append(selevel)

        default_secontext = self.selinux_default_context(path)
        for i in range(len(default_secontext)):
            if i is not None and secontext[i] == '_default':
                secontext[i] = default_secontext[i]

        attributes = params.get('attributes', None)
        return dict(
            path=path, mode=mode, owner=owner, group=group,
            seuser=seuser, serole=serole, setype=setype,
            selevel=selevel, secontext=secontext, attributes=attributes,
        )

    # Detect whether using selinux that is MLS-aware.
    # While this means you can set the level/range with
    # selinux.lsetfilecon(), it may or may not mean that you
    # will get the selevel as part of the context returned
    # by selinux.lgetfilecon().

    def selinux_mls_enabled(self):
        if self._selinux_mls_enabled is None:
            self._selinux_mls_enabled = HAVE_SELINUX and selinux.is_selinux_mls_enabled() == 1

        return self._selinux_mls_enabled

    def selinux_enabled(self):
        if self._selinux_enabled is None:
            self._selinux_enabled = HAVE_SELINUX and selinux.is_selinux_enabled() == 1

        return self._selinux_enabled

    # Determine whether we need a placeholder for selevel/mls
    def selinux_initial_context(self):
        if self._selinux_initial_context is None:
            self._selinux_initial_context = [None, None, None]
            if self.selinux_mls_enabled():
                self._selinux_initial_context.append(None)

        return self._selinux_initial_context

    # If selinux fails to find a default, return an array of None
    def selinux_default_context(self, path, mode=0):
        context = self.selinux_initial_context()
        if not self.selinux_enabled():
            return context
        try:
            ret = selinux.matchpathcon(to_native(path, errors='surrogate_or_strict'), mode)
        except OSError:
            return context
        if ret[0] == -1:
            return context
        # Limit split to 4 because the selevel, the last in the list,
        # may contain ':' characters
        context = ret[1].split(':', 3)
        return context

    def selinux_context(self, path):
        context = self.selinux_initial_context()
        if not self.selinux_enabled():
            return context
        try:
            ret = selinux.lgetfilecon_raw(to_native(path, errors='surrogate_or_strict'))
        except OSError as ex:
            self.fail_json(path=path, msg='Failed to retrieve selinux context.', exception=ex)
        if ret[0] == -1:
            return context
        # Limit split to 4 because the selevel, the last in the list,
        # may contain ':' characters
        context = ret[1].split(':', 3)
        return context

    def user_and_group(self, path, expand=True):
        b_path = to_bytes(path, errors='surrogate_or_strict')
        if expand:
            b_path = os.path.expanduser(os.path.expandvars(b_path))
        st = os.lstat(b_path)
        uid = st.st_uid
        gid = st.st_gid
        return (uid, gid)

    def find_mount_point(self, path):
        """
            Takes a path and returns its mount point

        :param path: a string type with a filesystem path
        :returns: the path to the mount point as a text type
        """

        b_path = os.path.realpath(to_bytes(os.path.expanduser(os.path.expandvars(path)), errors='surrogate_or_strict'))
        while not os.path.ismount(b_path):
            b_path = os.path.dirname(b_path)

        return to_text(b_path, errors='surrogate_or_strict')

    def is_special_selinux_path(self, path):
        """
        Returns a tuple containing (True, selinux_context) if the given path is on a
        NFS or other 'special' fs  mount point, otherwise the return will be (False, None).
        """
        try:
            with open('/proc/mounts', 'r') as f:
                mount_data = f.readlines()
        except Exception:
            return (False, None)

        path_mount_point = self.find_mount_point(path)

        for line in mount_data:
            (device, mount_point, fstype, options, rest) = line.split(' ', 4)
            if to_bytes(path_mount_point) == to_bytes(mount_point):
                for fs in self._selinux_special_fs:
                    if fs in fstype:
                        special_context = self.selinux_context(path_mount_point)
                        return (True, special_context)

        return (False, None)

    def set_default_selinux_context(self, path, changed):
        if not self.selinux_enabled():
            return changed
        context = self.selinux_default_context(path)
        return self.set_context_if_different(path, context, False)

    def set_context_if_different(self, path, context, changed, diff=None):

        if not self.selinux_enabled():
            return changed

        if self.check_file_absent_if_check_mode(path):
            return True

        cur_context = self.selinux_context(path)
        new_context = list(cur_context)
        # Iterate over the current context instead of the
        # argument context, which may have selevel.

        (is_special_se, sp_context) = self.is_special_selinux_path(path)
        if is_special_se:
            new_context = sp_context
        else:
            for i in range(len(cur_context)):
                if len(context) > i:
                    if context[i] is not None and context[i] != cur_context[i]:
                        new_context[i] = context[i]
                    elif context[i] is None:
                        new_context[i] = cur_context[i]

        if cur_context != new_context:
            if diff is not None:
                if 'before' not in diff:
                    diff['before'] = {}
                diff['before']['secontext'] = cur_context
                if 'after' not in diff:
                    diff['after'] = {}
                diff['after']['secontext'] = new_context

            try:
                if self.check_mode:
                    return True
                rc = selinux.lsetfilecon(to_native(path), ':'.join(new_context))
            except OSError as e:
                self.fail_json(path=path, msg='invalid selinux context: %s' % to_native(e),
                               new_context=new_context, cur_context=cur_context, input_was=context)
            if rc != 0:
                self.fail_json(path=path, msg='set selinux context failed')
            changed = True
        return changed

    def set_owner_if_different(self, path, owner, changed, diff=None, expand=True):

        if owner is None:
            return changed

        b_path = to_bytes(path, errors='surrogate_or_strict')
        if expand:
            b_path = os.path.expanduser(os.path.expandvars(b_path))

        if self.check_file_absent_if_check_mode(b_path):
            return True

        orig_uid, orig_gid = self.user_and_group(b_path, expand)
        try:
            uid = int(owner)
        except ValueError:
            try:
                uid = pwd.getpwnam(owner).pw_uid
            except KeyError:
                path = to_text(b_path)
                self.fail_json(path=path, msg='chown failed: failed to look up user %s' % owner)

        if orig_uid != uid:
            if diff is not None:
                if 'before' not in diff:
                    diff['before'] = {}
                diff['before']['owner'] = orig_uid
                if 'after' not in diff:
                    diff['after'] = {}
                diff['after']['owner'] = uid

            if self.check_mode:
                return True
            try:
                os.lchown(b_path, uid, -1)
            except OSError as ex:
                path = to_text(b_path)
                self.fail_json(path=path, msg='chown failed', exception=ex)
            changed = True
        return changed

    def set_group_if_different(self, path, group, changed, diff=None, expand=True):

        if group is None:
            return changed

        b_path = to_bytes(path, errors='surrogate_or_strict')
        if expand:
            b_path = os.path.expanduser(os.path.expandvars(b_path))

        if self.check_file_absent_if_check_mode(b_path):
            return True

        orig_uid, orig_gid = self.user_and_group(b_path, expand)
        try:
            gid = int(group)
        except ValueError:
            try:
                gid = grp.getgrnam(group).gr_gid
            except KeyError:
                path = to_text(b_path)
                self.fail_json(path=path, msg='chgrp failed: failed to look up group %s' % group)

        if orig_gid != gid:
            if diff is not None:
                if 'before' not in diff:
                    diff['before'] = {}
                diff['before']['group'] = orig_gid
                if 'after' not in diff:
                    diff['after'] = {}
                diff['after']['group'] = gid

            if self.check_mode:
                return True
            try:
                os.lchown(b_path, -1, gid)
            except OSError:
                path = to_text(b_path)
                self.fail_json(path=path, msg='chgrp failed')
            changed = True
        return changed

    def set_mode_if_different(self, path, mode, changed, diff=None, expand=True):

        if mode is None:
            return changed

        b_path = to_bytes(path, errors='surrogate_or_strict')
        if expand:
            b_path = os.path.expanduser(os.path.expandvars(b_path))

        if self.check_file_absent_if_check_mode(b_path):
            return True

        path_stat = os.lstat(b_path)

        if not isinstance(mode, int):
            try:
                mode = int(mode, 8)
            except Exception:
                try:
                    mode = self._symbolic_mode_to_octal(path_stat, mode)
                except Exception as e:
                    path = to_text(b_path)
                    self.fail_json(path=path,
                                   msg="mode must be in octal or symbolic form",
                                   details=to_native(e))

                if mode != stat.S_IMODE(mode):
                    # prevent mode from having extra info or being invalid long number
                    path = to_text(b_path)
                    self.fail_json(path=path, msg="Invalid mode supplied, only permission info is allowed", details=mode)

        prev_mode = stat.S_IMODE(path_stat.st_mode)

        if prev_mode != mode:

            if diff is not None:
                if 'before' not in diff:
                    diff['before'] = {}
                diff['before']['mode'] = '0%03o' % prev_mode
                if 'after' not in diff:
                    diff['after'] = {}
                diff['after']['mode'] = '0%03o' % mode

            if self.check_mode:
                return True
            # FIXME: comparison against string above will cause this to be executed
            # every time
            try:
                if hasattr(os, 'lchmod'):
                    os.lchmod(b_path, mode)
                else:
                    if not os.path.islink(b_path):
                        os.chmod(b_path, mode)
                    else:
                        # Attempt to set the perms of the symlink but be
                        # careful not to change the perms of the underlying
                        # file while trying
                        underlying_stat = os.stat(b_path)
                        os.chmod(b_path, mode)
                        new_underlying_stat = os.stat(b_path)
                        if underlying_stat.st_mode != new_underlying_stat.st_mode:
                            os.chmod(b_path, stat.S_IMODE(underlying_stat.st_mode))
            except OSError as e:
                if os.path.islink(b_path) and e.errno in (
                    errno.EACCES,  # can't access symlink in sticky directory (stat)
                    errno.EPERM,  # can't set mode on symbolic links (chmod)
                    errno.EROFS,  # can't set mode on read-only filesystem
                ):
                    pass
                elif e.errno in (errno.ENOENT, errno.ELOOP):  # Can't set mode on broken symbolic links
                    pass
                else:
                    raise
            except Exception as e:
                path = to_text(b_path)
                self.fail_json(path=path, msg='chmod failed', details=to_native(e))

            path_stat = os.lstat(b_path)
            new_mode = stat.S_IMODE(path_stat.st_mode)

            if new_mode != prev_mode:
                changed = True
        return changed

    def set_attributes_if_different(self, path, attributes, changed, diff=None, expand=True):

        if attributes is None:
            return changed

        b_path = to_bytes(path, errors='surrogate_or_strict')
        if expand:
            b_path = os.path.expanduser(os.path.expandvars(b_path))

        if self.check_file_absent_if_check_mode(b_path):
            return True

        existing = self.get_file_attributes(b_path, include_version=False)

        attr_mod = '='
        if attributes.startswith(('-', '+')):
            attr_mod = attributes[0]
            attributes = attributes[1:]

        if attributes and (existing.get('attr_flags', '') != attributes or attr_mod == '-'):
            attrcmd = self.get_bin_path('chattr')
            if attrcmd:
                attrcmd = [attrcmd, '%s%s' % (attr_mod, attributes), b_path]
                changed = True

                if diff is not None:
                    if 'before' not in diff:
                        diff['before'] = {}
                    diff['before']['attributes'] = existing.get('attr_flags')
                    if 'after' not in diff:
                        diff['after'] = {}
                    diff['after']['attributes'] = '%s%s' % (attr_mod, attributes)

                if not self.check_mode:
                    try:
                        rc, out, err = self.run_command(attrcmd)
                        if rc != 0 or err:
                            raise Exception("Error while setting attributes: %s" % (out + err))
                    except Exception as e:
                        self.fail_json(path=to_text(b_path), msg='chattr failed', details=to_native(e))
        return changed

    def get_file_attributes(self, path, include_version=True):
        output = {}
        attrcmd = self.get_bin_path('lsattr', False)
        if attrcmd:
            flags = '-vd' if include_version else '-d'
            attrcmd = [attrcmd, flags, path]
            try:
                rc, out, err = self.run_command(attrcmd)
                if rc == 0:
                    res = out.split()
                    attr_flags_idx = 0
                    if include_version:
                        attr_flags_idx = 1
                        output['version'] = res[0].strip()
                    output['attr_flags'] = res[attr_flags_idx].replace('-', '').strip()
                    output['attributes'] = format_attributes(output['attr_flags'])
            except Exception:
                pass
        return output

    @classmethod
    def _symbolic_mode_to_octal(cls, path_stat, symbolic_mode):
        """
        This enables symbolic chmod string parsing as stated in the chmod man-page

        This includes things like: "u=rw-x+X,g=r-x+X,o=r-x+X"
        """

        new_mode = stat.S_IMODE(path_stat.st_mode)

        # Now parse all symbolic modes
        for mode in symbolic_mode.split(','):
            # Per single mode. This always contains a '+', '-' or '='
            # Split it on that
            permlist = MODE_OPERATOR_RE.split(mode)

            # And find all the operators
            opers = MODE_OPERATOR_RE.findall(mode)

            # The user(s) where it's all about is the first element in the
            # 'permlist' list. Take that and remove it from the list.
            # An empty user or 'a' means 'all'.
            users = permlist.pop(0)
            use_umask = (users == '')
            if users == 'a' or users == '':
                users = 'ugo'

            # Check if there are illegal characters in the user list
            # They can end up in 'users' because they are not split
            if not USERS_RE.match(users):
                raise ValueError("bad symbolic permission for mode: %s" % mode)

            # Now we have two list of equal length, one contains the requested
            # permissions and one with the corresponding operators.
            for idx, perms in enumerate(permlist):
                # Check if there are illegal characters in the permissions
                if not PERMS_RE.match(perms):
                    raise ValueError("bad symbolic permission for mode: %s" % mode)

                for user in users:
                    mode_to_apply = cls._get_octal_mode_from_symbolic_perms(path_stat, user, perms, use_umask, new_mode)
                    new_mode = cls._apply_operation_to_mode(user, opers[idx], mode_to_apply, new_mode)

        return new_mode

    @staticmethod
    def _apply_operation_to_mode(user, operator, mode_to_apply, current_mode):
        if operator == '=':
            if user == 'u':
                mask = stat.S_IRWXU | stat.S_ISUID
            elif user == 'g':
                mask = stat.S_IRWXG | stat.S_ISGID
            elif user == 'o':
                mask = stat.S_IRWXO | stat.S_ISVTX

            # mask out u, g, or o permissions from current_mode and apply new permissions
            inverse_mask = mask ^ PERM_BITS
            new_mode = (current_mode & inverse_mask) | mode_to_apply
        elif operator == '+':
            new_mode = current_mode | mode_to_apply
        elif operator == '-':
            new_mode = current_mode - (current_mode & mode_to_apply)
        return new_mode

    @staticmethod
    def _get_octal_mode_from_symbolic_perms(path_stat, user, perms, use_umask, prev_mode=None):
        if prev_mode is None:
            prev_mode = stat.S_IMODE(path_stat.st_mode)
        is_directory = stat.S_ISDIR(path_stat.st_mode)
        has_x_permissions = (prev_mode & S_IXANY) > 0
        apply_X_permission = is_directory or has_x_permissions

        # Get the umask, if the 'user' part is empty, the effect is as if (a) were
        # given, but bits that are set in the umask are not affected.
        # We also need the "reversed umask" for masking
        umask = os.umask(0)
        os.umask(umask)
        rev_umask = umask ^ PERM_BITS

        # Permission bits constants documented at:
        # https://docs.python.org/3/library/stat.html#stat.S_ISUID
        if apply_X_permission:
            X_perms = {
                'u': {'X': stat.S_IXUSR},
                'g': {'X': stat.S_IXGRP},
                'o': {'X': stat.S_IXOTH},
            }
        else:
            X_perms = {
                'u': {'X': 0},
                'g': {'X': 0},
                'o': {'X': 0},
            }

        user_perms_to_modes = {
            'u': {
                'r': rev_umask & stat.S_IRUSR if use_umask else stat.S_IRUSR,
                'w': rev_umask & stat.S_IWUSR if use_umask else stat.S_IWUSR,
                'x': rev_umask & stat.S_IXUSR if use_umask else stat.S_IXUSR,
                's': stat.S_ISUID,
                't': 0,
                'u': prev_mode & stat.S_IRWXU,
                'g': (prev_mode & stat.S_IRWXG) << 3,
                'o': (prev_mode & stat.S_IRWXO) << 6},
            'g': {
                'r': rev_umask & stat.S_IRGRP if use_umask else stat.S_IRGRP,
                'w': rev_umask & stat.S_IWGRP if use_umask else stat.S_IWGRP,
                'x': rev_umask & stat.S_IXGRP if use_umask else stat.S_IXGRP,
                's': stat.S_ISGID,
                't': 0,
                'u': (prev_mode & stat.S_IRWXU) >> 3,
                'g': prev_mode & stat.S_IRWXG,
                'o': (prev_mode & stat.S_IRWXO) << 3},
            'o': {
                'r': rev_umask & stat.S_IROTH if use_umask else stat.S_IROTH,
                'w': rev_umask & stat.S_IWOTH if use_umask else stat.S_IWOTH,
                'x': rev_umask & stat.S_IXOTH if use_umask else stat.S_IXOTH,
                's': 0,
                't': stat.S_ISVTX,
                'u': (prev_mode & stat.S_IRWXU) >> 6,
                'g': (prev_mode & stat.S_IRWXG) >> 3,
                'o': prev_mode & stat.S_IRWXO},
        }

        # Insert X_perms into user_perms_to_modes
        for key, value in X_perms.items():
            user_perms_to_modes[key].update(value)

        def or_reduce(mode, perm):
            return mode | user_perms_to_modes[user][perm]

        return reduce(or_reduce, perms, 0)

    def set_fs_attributes_if_different(self, file_args, changed, diff=None, expand=True):
        # set modes owners and context as needed
        changed = self.set_context_if_different(
            file_args['path'], file_args['secontext'], changed, diff
        )
        changed = self.set_owner_if_different(
            file_args['path'], file_args['owner'], changed, diff, expand
        )
        changed = self.set_group_if_different(
            file_args['path'], file_args['group'], changed, diff, expand
        )
        changed = self.set_mode_if_different(
            file_args['path'], file_args['mode'], changed, diff, expand
        )
        changed = self.set_attributes_if_different(
            file_args['path'], file_args['attributes'], changed, diff, expand
        )
        return changed

    def check_file_absent_if_check_mode(self, file_path):
        return self.check_mode and not os.path.exists(file_path)

    def set_directory_attributes_if_different(self, file_args, changed, diff=None, expand=True):
        return self.set_fs_attributes_if_different(file_args, changed, diff, expand)

    def set_file_attributes_if_different(self, file_args, changed, diff=None, expand=True):
        return self.set_fs_attributes_if_different(file_args, changed, diff, expand)

    def add_path_info(self, kwargs):
        """
        for results that are files, supplement the info about the file
        in the return path with stats about the file path.
        """

        path = kwargs.get('path', kwargs.get('dest', None))
        if path is None:
            return kwargs
        b_path = to_bytes(path, errors='surrogate_or_strict')
        if os.path.exists(b_path):
            (uid, gid) = self.user_and_group(path)
            kwargs['uid'] = uid
            kwargs['gid'] = gid
            try:
                user = pwd.getpwuid(uid)[0]
            except KeyError:
                user = str(uid)
            try:
                group = grp.getgrgid(gid)[0]
            except KeyError:
                group = str(gid)
            kwargs['owner'] = user
            kwargs['group'] = group
            st = os.lstat(b_path)
            kwargs['mode'] = '0%03o' % stat.S_IMODE(st[stat.ST_MODE])
            # secontext not yet supported
            if os.path.islink(b_path):
                kwargs['state'] = 'link'
            elif os.path.isdir(b_path):
                kwargs['state'] = 'directory'
            elif os.stat(b_path).st_nlink > 1:
                kwargs['state'] = 'hard'
            else:
                kwargs['state'] = 'file'
            if self.selinux_enabled():
                kwargs['secontext'] = ':'.join(self.selinux_context(path))
            kwargs['size'] = st[stat.ST_SIZE]
        return kwargs

    def _check_locale(self):
        """
        Uses the locale module to test the currently set locale
        (per the LANG and LC_CTYPE environment settings)
        """
        try:
            # setting the locale to '' uses the default locale
            # as it would be returned by locale.getdefaultlocale()
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            # fallback to the 'best' locale, per the function
            # final fallback is 'C', which may cause unicode issues
            # but is preferable to simply failing on unknown locale
            best_locale = get_best_parsable_locale(self)

            # need to set several since many tools choose to ignore documented precedence and scope
            locale.setlocale(locale.LC_ALL, best_locale)
            os.environ['LANG'] = best_locale
            os.environ['LC_ALL'] = best_locale
            os.environ['LC_MESSAGES'] = best_locale
        except Exception as e:
            self.fail_json(msg="An unknown error was encountered while attempting to validate the locale: %s" % to_native(e))

    def _set_internal_properties(self, argument_spec=None, module_parameters=None):
        if argument_spec is None:
            argument_spec = self.argument_spec
        if module_parameters is None:
            module_parameters = self.params

        for k in PASS_VARS:
            # handle setting internal properties from internal ansible vars
            param_key = '_ansible_%s' % k
            if param_key in module_parameters:
                if k in PASS_BOOLS:
                    setattr(self, PASS_VARS[k][0], self.boolean(module_parameters[param_key]))
                else:
                    setattr(self, PASS_VARS[k][0], module_parameters[param_key])

                # clean up internal top level params:
                if param_key in self.params:
                    del self.params[param_key]
            else:
                # use defaults if not already set
                if not hasattr(self, PASS_VARS[k][0]):
                    setattr(self, PASS_VARS[k][0], PASS_VARS[k][1])

    def safe_eval(self, value, locals=None, include_exceptions=False):
        # deprecated: description='no longer used in the codebase' core_version='2.21'
        return safe_eval(value, locals, include_exceptions)

    def _load_params(self):
        """ read the input and set the params attribute.

        This method is for backwards compatibility.  The guts of the function
        were moved out in 2.1 so that custom modules could read the parameters.
        """
        # debug overrides to read args from file or cmdline
        self.params = _load_params()

    def _log_to_syslog(self, msg):
        if HAS_SYSLOG:
            try:
                module = 'ansible-%s' % self._name
                facility = getattr(syslog, self._syslog_facility, syslog.LOG_USER)
                syslog.openlog(str(module), 0, facility)
                syslog.syslog(syslog.LOG_INFO, msg)
            except (TypeError, ValueError) as e:
                self.fail_json(
                    msg='Failed to log to syslog (%s). To proceed anyway, '
                        'disable syslog logging by setting no_target_syslog '
                        'to True in your Ansible config.' % to_native(e),
                    msg_to_log=msg,
                )

    def debug(self, msg):
        if self._debug:
            self.log('[debug] %s' % msg)

    def log(self, msg, log_args=None):

        if not self.no_log:

            if log_args is None:
                log_args = dict()

            module = 'ansible-%s' % self._name
            if isinstance(module, bytes):
                module = module.decode('utf-8', 'replace')

            # 6655 - allow for accented characters
            if not isinstance(msg, (bytes, str)):
                raise TypeError("msg should be a string (got %s)" % type(msg))

            # We want journal to always take text type
            # syslog takes bytes on py2, text type on py3
            if isinstance(msg, bytes):
                journal_msg = msg.decode('utf-8', 'replace')
            else:
                # TODO: surrogateescape is a danger here on Py3
                journal_msg = msg

            if self._target_log_info:
                journal_msg = ' '.join([self._target_log_info, journal_msg])

            # ensure we clean up secrets!
            journal_msg = remove_values(journal_msg, self.no_log_values)

            if has_journal:
                journal_args = [("MODULE", os.path.basename(__file__))]
                for arg in log_args:
                    name, value = (arg.upper(), str(log_args[arg]))
                    if name in (
                        'PRIORITY', 'MESSAGE', 'MESSAGE_ID',
                        'CODE_FILE', 'CODE_LINE', 'CODE_FUNC',
                        'SYSLOG_FACILITY', 'SYSLOG_IDENTIFIER',
                        'SYSLOG_PID',
                    ):
                        name = "_%s" % name
                    journal_args.append((name, value))

                try:
                    if HAS_SYSLOG:
                        # If syslog_facility specified, it needs to convert
                        #  from the facility name to the facility code, and
                        #  set it as SYSLOG_FACILITY argument of journal.send()
                        facility = getattr(syslog,
                                           self._syslog_facility,
                                           syslog.LOG_USER) >> 3
                        journal.send(MESSAGE=u"%s %s" % (module, journal_msg),
                                     SYSLOG_FACILITY=facility,
                                     **dict(journal_args))
                    else:
                        journal.send(MESSAGE=u"%s %s" % (module, journal_msg),
                                     **dict(journal_args))
                except OSError:
                    # fall back to syslog since logging to journal failed
                    self._log_to_syslog(journal_msg)
            else:
                self._log_to_syslog(journal_msg)

    def _log_invocation(self):
        """ log that ansible ran the module """
        # TODO: generalize a separate log function and make log_invocation use it
        # Sanitize possible password argument when logging.
        log_args = dict()

        for param in self.params:
            canon = self.aliases.get(param, param)
            arg_opts = self.argument_spec.get(canon, {})
            no_log = arg_opts.get('no_log', None)

            # try to proactively capture password/passphrase fields
            if no_log is None and PASSWORD_MATCH.search(param):
                log_args[param] = 'NOT_LOGGING_PASSWORD'
                self.warn('Module did not set no_log for %s' % param)
            elif self.boolean(no_log):
                log_args[param] = 'NOT_LOGGING_PARAMETER'
            else:
                param_val = self.params[param]
                if not isinstance(param_val, (str, bytes)):
                    param_val = str(param_val)
                elif isinstance(param_val, str):
                    param_val = param_val.encode('utf-8')
                log_args[param] = heuristic_log_sanitize(param_val, self.no_log_values)

        msg = ['%s=%s' % (to_native(arg), to_native(val)) for arg, val in log_args.items()]
        if msg:
            msg = 'Invoked with %s' % ' '.join(msg)
        else:
            msg = 'Invoked'

        self.log(msg, log_args=log_args)

    def _set_cwd(self):
        try:
            cwd = os.getcwd()
            if not os.access(cwd, os.F_OK | os.R_OK):
                raise Exception()
            return cwd
        except Exception:
            # we don't have access to the cwd, probably because of sudo.
            # Try and move to a neutral location to prevent errors
            for cwd in [self.tmpdir, os.path.expandvars('$HOME'), tempfile.gettempdir()]:
                try:
                    if os.access(cwd, os.F_OK | os.R_OK):
                        os.chdir(cwd)
                        return cwd
                except Exception:
                    pass
        # we won't error here, as it may *not* be a problem,
        # and we don't want to break modules unnecessarily
        return None

    def get_bin_path(self, arg, required=False, opt_dirs=None):
        """
        Find system executable in PATH.

        :param arg: The executable to find.
        :param required: if the executable is not found and required is ``True``, fail_json
        :param opt_dirs: optional list of directories to search in addition to ``PATH``
        :returns: if found return full path; otherwise return original arg, unless 'warning' then return None
        :raises: Sysexit: if arg is not found and required=True (via fail_json)
        """

        bin_path = None
        try:
            bin_path = get_bin_path(arg=arg, opt_dirs=opt_dirs)
        except ValueError as e:
            if required:
                self.fail_json(msg=to_text(e))

        return bin_path

    def boolean(self, arg):
        """Convert the argument to a boolean"""
        if arg is None:
            return arg

        try:
            return boolean(arg)
        except TypeError as e:
            self.fail_json(msg=to_native(e))

    def jsonify(self, data):
        # deprecated: description='deprecate AnsibleModule.jsonify()' core_version='2.23'
        # deprecate(
        #     msg="The `AnsibleModule.jsonify' method is deprecated.",
        #     version="2.27",
        #     # help_text="",  # DTFIX-FUTURE: fill in this help text
        # )

        try:
            return json.dumps(data, cls=_common_json._get_legacy_encoder())
        except UnicodeError as e:
            self.fail_json(msg=to_text(e))

    def from_json(self, data):
        return json.loads(data)

    def add_cleanup_file(self, path):
        if path not in self.cleanup_files:
            self.cleanup_files.append(path)

    def do_cleanup_files(self):
        for path in self.cleanup_files:
            self.cleanup(path)

    def _return_formatted(self, kwargs):
        _skip_stackwalk = True

        self.add_path_info(kwargs)

        if 'invocation' not in kwargs:
            kwargs['invocation'] = {'module_args': self.params}

        if 'warnings' in kwargs:
            self.deprecate(  # pylint: disable=ansible-deprecated-unnecessary-collection-name
                msg='Passing `warnings` to `exit_json` or `fail_json` is deprecated.',
                version='2.23',
                help_text='Use `AnsibleModule.warn` instead.',
                deprecator=_deprecator.ANSIBLE_CORE_DEPRECATOR,
            )

            if isinstance(kwargs['warnings'], list):
                for w in kwargs['warnings']:
                    self.warn(w)
            else:
                self.warn(kwargs['warnings'])

        warnings = get_warnings()
        if warnings:
            kwargs['warnings'] = warnings

        if 'deprecations' in kwargs:
            self.deprecate(  # pylint: disable=ansible-deprecated-unnecessary-collection-name
                msg='Passing `deprecations` to `exit_json` or `fail_json` is deprecated.',
                version='2.23',
                help_text='Use `AnsibleModule.deprecate` instead.',
                deprecator=_deprecator.ANSIBLE_CORE_DEPRECATOR,
            )

            if isinstance(kwargs['deprecations'], list):
                for d in kwargs['deprecations']:
                    if isinstance(d, (KeysView, Sequence)) and len(d) == 2:
                        self.deprecate(  # pylint: disable=ansible-deprecated-unnecessary-collection-name,ansible-invalid-deprecated-version
                            msg=d[0],
                            version=d[1],
                            deprecator=_deprecator.get_best_deprecator(),
                        )
                    elif isinstance(d, Mapping):
                        self.deprecate(  # pylint: disable=ansible-deprecated-date-not-permitted,ansible-deprecated-unnecessary-collection-name
                            msg=d['msg'],
                            version=d.get('version'),
                            date=d.get('date'),
                            deprecator=_deprecator.get_best_deprecator(collection_name=d.get('collection_name')),
                        )
                    else:
                        self.deprecate(  # pylint: disable=ansible-deprecated-unnecessary-collection-name,ansible-deprecated-no-version
                            msg=d,
                            deprecator=_deprecator.get_best_deprecator(),
                        )
            else:
                self.deprecate(  # pylint: disable=ansible-deprecated-unnecessary-collection-name,ansible-deprecated-no-version
                    msg=kwargs['deprecations'],
                    deprecator=_deprecator.get_best_deprecator(),
                )

        deprecations = get_deprecations()
        if deprecations:
            kwargs['deprecations'] = deprecations

        # preserve bools/none from no_log
        preserved = {k: v for k, v in kwargs.items() if v is None or isinstance(v, bool)}

        # strip no_log collisions
        kwargs = remove_values(kwargs, self.no_log_values)

        # graft preserved values back on
        kwargs.update(preserved)

        self._record_module_result(kwargs)

    def _record_module_result(self, o: dict[str, t.Any]) -> None:
        """
        Temporary internal hook to enable modification/bypass of module result serialization.

        Monkeypatched by ansible.netcommon for direct in-worker module execution.
        """
        encoder = _json.get_module_encoder(_ANSIBLE_PROFILE, _json.Direction.MODULE_TO_CONTROLLER)
        print('\n%s' % json.dumps(o, cls=encoder))

    def exit_json(self, **kwargs) -> t.NoReturn:
        """ return from the module, without error """
        _skip_stackwalk = True

        self.do_cleanup_files()
        self._return_formatted(kwargs)
        sys.exit(0)

    def fail_json(self, msg: str, *, exception: BaseException | str | None = _UNSET, **kwargs) -> t.NoReturn:
        """
        Return from the module with an error message and optional exception/traceback detail.
        A traceback will only be included in the result if error traceback capturing has been enabled.

        When `exception` is an exception object, its message chain will be automatically combined with `msg` to create the final error message.
        The message chain includes the exception's message as well as messages from any __cause__ exceptions.
        The traceback from `exception` will be used for the formatted traceback.

        When `exception` is a string, it will be used as the formatted traceback.

        When `exception` is set to `None`, the current call stack will be used for the formatted traceback.

        When `exception` is not specified, a formatted traceback will be retrieved from the current exception.
        If no exception is pending, the current call stack will be used instead.
        """
        _skip_stackwalk = True

        msg = str(msg)  # coerce to str instead of raising an error due to an invalid type

        kwargs.update(
            failed=True,
            msg=msg,
        )

        if isinstance(exception, BaseException):
            # Include a `_messages.Event` in the result.
            # The `msg` is included in the chain to ensure it is not lost when looking only at `exception` from the result.

            kwargs.update(
                exception=_messages.ErrorSummary(
                    event=_messages.Event(
                        msg=msg,
                        formatted_traceback=_traceback.maybe_capture_traceback(msg, _traceback.TracebackEvent.ERROR),
                        chain=_messages.EventChain(
                            msg_reason=_errors.MSG_REASON_DIRECT_CAUSE,
                            traceback_reason="The above exception was the direct cause of the following error:",
                            event=_errors.EventFactory.from_exception(exception, _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR)),
                        ),
                    ),
                ),
            )
        elif _traceback.is_traceback_enabled(_traceback.TracebackEvent.ERROR):
            # Include only a formatted traceback string in the result.
            # The controller will combine this with `msg` to create an `_messages.ErrorSummary`.

            formatted_traceback: str | None

            if isinstance(exception, str):
                formatted_traceback = exception
            elif exception is _UNSET and (current_exception := t.cast(t.Optional[BaseException], sys.exc_info()[1])):
                formatted_traceback = _traceback.maybe_extract_traceback(current_exception, _traceback.TracebackEvent.ERROR)
            else:
                formatted_traceback = _traceback.maybe_capture_traceback(msg, _traceback.TracebackEvent.ERROR)

            if formatted_traceback:
                kwargs.update(exception=formatted_traceback)

        self.do_cleanup_files()
        self._return_formatted(kwargs)

        sys.exit(1)

    def fail_on_missing_params(self, required_params=None):
        if not required_params:
            return
        try:
            check_missing_parameters(self.params, required_params)
        except TypeError as e:
            self.fail_json(msg=to_native(e))

    def digest_from_file(self, filename, algorithm):
        """ Return hex digest of local file for a digest_method specified by name, or None if file is not present. """
        b_filename = to_bytes(filename, errors='surrogate_or_strict')

        if not os.path.exists(b_filename):
            return None
        if os.path.isdir(b_filename):
            self.fail_json(msg="attempted to take checksum of directory: %s" % filename)

        # preserve old behaviour where the third parameter was a hash algorithm object
        if hasattr(algorithm, 'hexdigest'):
            digest_method = algorithm
        else:
            try:
                digest_method = AVAILABLE_HASH_ALGORITHMS[algorithm]()
            except KeyError:
                self.fail_json(msg="Could not hash file '%s' with algorithm '%s'. Available algorithms: %s" %
                                   (filename, algorithm, ', '.join(AVAILABLE_HASH_ALGORITHMS)))

        blocksize = 64 * 1024
        infile = open(os.path.realpath(b_filename), 'rb')
        block = infile.read(blocksize)
        while block:
            digest_method.update(block)
            block = infile.read(blocksize)
        infile.close()
        return digest_method.hexdigest()

    def md5(self, filename):
        """ Return MD5 hex digest of local file using digest_from_file().

        Do not use this function unless you have no other choice for:
            1) Optional backwards compatibility
            2) Compatibility with a third party protocol

        This function will not work on systems complying with FIPS-140-2.

        Most uses of this function can use the module.sha1 function instead.
        """
        if 'md5' not in AVAILABLE_HASH_ALGORITHMS:
            raise ValueError('MD5 not available.  Possibly running in FIPS mode')
        return self.digest_from_file(filename, 'md5')

    def sha1(self, filename):
        """ Return SHA1 hex digest of local file using digest_from_file(). """
        return self.digest_from_file(filename, 'sha1')

    def sha256(self, filename):
        """ Return SHA-256 hex digest of local file using digest_from_file(). """
        return self.digest_from_file(filename, 'sha256')

    def backup_local(self, fn):
        """make a date-marked backup of the specified file, return True or False on success or failure"""

        backupdest = ''
        if os.path.exists(fn):
            # backups named basename.PID.YYYY-MM-DD@HH:MM:SS~
            ext = time.strftime("%Y-%m-%d@%H:%M:%S~", time.localtime(time.time()))
            backupdest = '%s.%s.%s' % (fn, os.getpid(), ext)

            if not self.check_mode:
                try:
                    self.preserved_copy(fn, backupdest)
                except (shutil.Error, IOError) as ex:
                    raise Exception(f'Could not make backup of {fn!r} to {backupdest!r}.') from ex

        return backupdest

    def cleanup(self, tmpfile):
        if os.path.exists(tmpfile):
            try:
                os.unlink(tmpfile)
            except OSError as e:
                sys.stderr.write("could not cleanup %s: %s" % (tmpfile, to_native(e)))

    def preserved_copy(self, src, dest):
        """Copy a file with preserved ownership, permissions and context"""

        # shutil.copy2(src, dst)
        #   Similar to shutil.copy(), but metadata is copied as well - in fact,
        #   this is just shutil.copy() followed by copystat(). This is similar
        #   to the Unix command cp -p.

        # shutil.copystat(src, dst)
        #   Copy the permission bits, last access time, last modification time,
        #   and flags from src to dst. The file contents, owner, and group are
        #   unaffected. src and dst are path names given as strings.

        shutil.copy2(src, dest)

        # Set the context
        if self.selinux_enabled():
            context = self.selinux_context(src)
            self.set_context_if_different(dest, context, False)

        # chown it
        try:
            dest_stat = os.stat(src)
            tmp_stat = os.stat(dest)
            if dest_stat and (tmp_stat.st_uid != dest_stat.st_uid or tmp_stat.st_gid != dest_stat.st_gid):
                os.chown(dest, dest_stat.st_uid, dest_stat.st_gid)
        except OSError as e:
            if e.errno != errno.EPERM:
                raise

        # Set the attributes
        current_attribs = self.get_file_attributes(src, include_version=False)
        current_attribs = current_attribs.get('attr_flags', '')
        self.set_attributes_if_different(dest, current_attribs, True)

    def atomic_move(self, src, dest, unsafe_writes=False, keep_dest_attrs=True):
        """atomically move src to dest, copying attributes from dest, returns true on success
        it uses os.rename to ensure this as it is an atomic operation, rest of the function is
        to work around limitations, corner cases and ensure selinux context is saved if possible"""
        context = None
        dest_stat = None
        b_src = to_bytes(src, errors='surrogate_or_strict')
        b_dest = to_bytes(dest, errors='surrogate_or_strict')
        if os.path.exists(b_dest) and keep_dest_attrs:
            try:
                dest_stat = os.stat(b_dest)
                os.chown(b_src, dest_stat.st_uid, dest_stat.st_gid)
                shutil.copystat(b_dest, b_src)
                os.utime(b_src, times=(time.time(), time.time()))
            except OSError as e:
                if e.errno != errno.EPERM:
                    raise
            if self.selinux_enabled():
                context = self.selinux_context(dest)
        else:
            if self.selinux_enabled():
                context = self.selinux_default_context(dest)

        creating = not os.path.exists(b_dest)

        try:
            # Optimistically try a rename, solves some corner cases and can avoid useless work, throws exception if not atomic.
            os.rename(b_src, b_dest)
        except OSError as ex:
            if ex.errno in (errno.EPERM, errno.EXDEV, errno.EACCES, errno.ETXTBSY, errno.EBUSY):
                # only try workarounds for errno 18 (cross device), 1 (not permitted),  13 (permission denied)
                # and 26 (text file busy) which happens on vagrant synced folders and other 'exotic' non posix file systems
                # Use bytes here.  In the shippable CI, this fails with
                # a UnicodeError with surrogateescape'd strings for an unknown
                # reason (doesn't happen in a local Ubuntu16.04 VM)
                b_dest_dir = os.path.dirname(b_dest)
                b_suffix = os.path.basename(b_dest)
                tmp_dest_name = None
                try:
                    tmp_dest_fd, tmp_dest_name = tempfile.mkstemp(prefix=b'.ansible_tmp', dir=b_dest_dir, suffix=b_suffix)
                except OSError as ex:
                    if unsafe_writes:
                        self._unsafe_writes(b_src, b_dest)
                    else:
                        raise Exception(
                            f'The destination directory {os.path.dirname(dest)!r} is not writable by the current user.'
                        ) from ex

                if tmp_dest_name:
                    b_tmp_dest_name = to_bytes(tmp_dest_name, errors='surrogate_or_strict')

                    try:
                        try:
                            # close tmp file handle before file operations to prevent text file busy errors on vboxfs synced folders (windows host)
                            os.close(tmp_dest_fd)
                            # leaves tmp file behind when sudo and not root
                            try:
                                shutil.move(b_src, b_tmp_dest_name, copy_function=shutil.copy if keep_dest_attrs else shutil.copy2)
                            except OSError:
                                # cleanup will happen by 'rm' of tmpdir
                                # copy2 will preserve some metadata
                                if keep_dest_attrs:
                                    shutil.copy(b_src, b_tmp_dest_name)
                                else:
                                    shutil.copy2(b_src, b_tmp_dest_name)

                            if self.selinux_enabled():
                                self.set_context_if_different(
                                    b_tmp_dest_name, context, False)
                            try:
                                tmp_stat = os.stat(b_tmp_dest_name)
                                if keep_dest_attrs:
                                    if dest_stat and (tmp_stat.st_uid != dest_stat.st_uid or tmp_stat.st_gid != dest_stat.st_gid):
                                        os.chown(b_tmp_dest_name, dest_stat.st_uid, dest_stat.st_gid)
                                    os.utime(b_tmp_dest_name, times=(time.time(), time.time()))
                            except OSError as ex:
                                if ex.errno != errno.EPERM:
                                    raise
                            try:
                                os.rename(b_tmp_dest_name, b_dest)
                            except (shutil.Error, OSError) as ex:
                                if unsafe_writes and ex.errno == errno.EBUSY:
                                    self._unsafe_writes(b_tmp_dest_name, b_dest)
                                else:
                                    raise Exception(
                                        f'Unable to make {src!r} into to {dest!r}, failed final rename from {to_text(b_tmp_dest_name)!r}.'
                                    ) from ex
                        except (shutil.Error, OSError) as ex:
                            if unsafe_writes:
                                self._unsafe_writes(b_src, b_dest)
                            else:
                                raise Exception(f'Failed to replace {dest!r} with {src!r}.') from ex
                    finally:
                        self.cleanup(b_tmp_dest_name)
            else:
                raise Exception(f'Could not replace {dest!r} with {src!r}.') from ex

        if creating:
            # make sure the file has the correct permissions
            # based on the current value of umask
            umask = os.umask(0)
            os.umask(umask)
            os.chmod(b_dest, S_IRWU_RWG_RWO & ~umask)
            dest_dir_stat = os.stat(os.path.dirname(b_dest))
            try:
                if dest_dir_stat.st_mode & stat.S_ISGID:
                    os.chown(b_dest, os.geteuid(), dest_dir_stat.st_gid)
                else:
                    os.chown(b_dest, os.geteuid(), os.getegid())
            except OSError:
                # We're okay with trying our best here.  If the user is not
                # root (or old Unices) they won't be able to chown.
                pass

        if self.selinux_enabled():
            # rename might not preserve context
            self.set_context_if_different(dest, context, False)

    def _unsafe_writes(self, src, dest):
        # sadly there are some situations where we cannot ensure atomicity, but only if
        # the user insists and we get the appropriate error we update the file unsafely
        try:
            with open(dest, 'wb') as out_dest:
                with open(src, 'rb') as in_src:
                    shutil.copyfileobj(in_src, out_dest)
        except (shutil.Error, OSError) as ex:
            raise Exception(f'Could not write data to file {dest!r} from {src!r}.') from ex

    def _clean_args(self, args):

        if not self._clean:
            # create a printable version of the command for use in reporting later,
            # which strips out things like passwords from the args list
            to_clean_args = args
            if isinstance(args, bytes):
                to_clean_args = to_text(args)
            if isinstance(args, (str, bytes)):
                to_clean_args = shlex.split(to_clean_args)

            clean_args = []
            is_passwd = False
            for arg in (to_native(a) for a in to_clean_args):
                if is_passwd:
                    is_passwd = False
                    clean_args.append('********')
                    continue
                if PASSWD_ARG_RE.match(arg):
                    sep_idx = arg.find('=')
                    if sep_idx > -1:
                        clean_args.append('%s=********' % arg[:sep_idx])
                        continue
                    else:
                        is_passwd = True
                arg = heuristic_log_sanitize(arg, self.no_log_values)
                clean_args.append(arg)
            self._clean = ' '.join(shlex.quote(arg) for arg in clean_args)

        return self._clean

    def run_command(self, args, check_rc=False, close_fds=True, executable=None, data=None, binary_data=False, path_prefix=None, cwd=None,
                    use_unsafe_shell=False, prompt_regex=None, environ_update=None, umask=None, encoding='utf-8', errors='surrogate_or_strict',
                    expand_user_and_vars=True, pass_fds=None, before_communicate_callback=None, ignore_invalid_cwd=True, handle_exceptions=True):
        """
        Execute a command, returns rc, stdout, and stderr.

        The mechanism of this method for reading stdout and stderr differs from
        that of CPython subprocess.Popen.communicate, in that this method will
        stop reading once the spawned command has exited and stdout and stderr
        have been consumed, as opposed to waiting until stdout/stderr are
        closed. This can be an important distinction, when taken into account
        that a forked or backgrounded process may hold stdout or stderr open
        for longer than the spawned command.

        :arg args: is the command to run
            * If args is a list, the command will be run with shell=False.
            * If args is a string and use_unsafe_shell=False it will split args to a list and run with shell=False
            * If args is a string and use_unsafe_shell=True it runs with shell=True.
        :kw check_rc: Whether to call fail_json in case of non zero RC.
            Default False
        :kw close_fds: See documentation for subprocess.Popen(). Default True
        :kw executable: See documentation for subprocess.Popen(). Default None
        :kw data: If given, information to write to the stdin of the command
        :kw binary_data: If False, append a newline to the data.  Default False
        :kw path_prefix: If given, additional path to find the command in.
            This adds to the PATH environment variable so helper commands in
            the same directory can also be found
        :kw cwd: If given, working directory to run the command inside
        :kw use_unsafe_shell: See `args` parameter.  Default False
        :kw prompt_regex: Regex string (not a compiled regex) which can be
            used to detect prompts in the stdout which would otherwise cause
            the execution to hang (especially if no input data is specified)
        :kw environ_update: dictionary to *update* environ variables with
        :kw umask: Umask to be used when running the command. Default None
        :kw encoding: Since we return strings, we need to
            know the encoding to use to transform from bytes to text.  If you
            want to always get bytes back, use encoding=None.  The default is
            "utf-8".  This does not affect transformation of strings given as
            args.
        :kw errors: Since we return strings, we need to
            transform stdout and stderr from bytes to text.  If the bytes are
            undecodable in the ``encoding`` specified, then use this error
            handler to deal with them.  The default is ``surrogate_or_strict``
            which means that the bytes will be decoded using the
            surrogateescape error handler if available (available on all
            Python versions we support) otherwise a UnicodeError traceback
            will be raised.  This does not affect transformations of strings
            given as args.
        :kw expand_user_and_vars: When ``use_unsafe_shell=False`` this argument
            dictates whether ``~`` is expanded in paths and environment variables
            are expanded before running the command. When ``True`` a string such as
            ``$SHELL`` will be expanded regardless of escaping. When ``False`` and
            ``use_unsafe_shell=False`` no path or variable expansion will be done.
        :kw pass_fds: This argument dictates which file descriptors should be passed
            to an underlying ``Popen`` constructor.
        :kw before_communicate_callback: This function will be called
            after ``Popen`` object will be created
            but before communicating to the process.
            (``Popen`` object will be passed to callback as a first argument)
        :kw ignore_invalid_cwd: This flag indicates whether an invalid ``cwd``
            (non-existent or not a directory) should be ignored or should raise
            an exception.
        :kw handle_exceptions: This flag indicates whether an exception will
            be handled inline and issue a failed_json or if the caller should
            handle it.
        :returns: A 3-tuple of return code (int), stdout (str), and stderr (str).
            stdout and stderr are text strings converted according to the encoding
            and errors parameters.  If you want byte strings, use encoding=None
            to turn decoding to text off.
        """
        # used by clean args later on
        self._clean = None

        if not isinstance(args, (list, bytes, str)):
            msg = "Argument 'args' to run_command must be list or string"
            self.fail_json(rc=257, cmd=args, msg=msg)

        shell = False
        if use_unsafe_shell:

            # stringify args for unsafe/direct shell usage
            if isinstance(args, list):
                args = b" ".join([to_bytes(shlex.quote(x), errors='surrogate_or_strict') for x in args])
            else:
                args = to_bytes(args, errors='surrogate_or_strict')

            # not set explicitly, check if set by controller
            if executable:
                executable = to_bytes(executable, errors='surrogate_or_strict')
                args = [executable, b'-c', args]
            elif self._shell not in (None, '/bin/sh'):
                args = [to_bytes(self._shell, errors='surrogate_or_strict'), b'-c', args]
            else:
                shell = True
        else:
            # ensure args are a list
            if isinstance(args, (bytes, str)):
                args = shlex.split(to_text(args, errors='surrogateescape'))

            # expand ``~`` in paths, and all environment vars
            if expand_user_and_vars:
                args = [to_bytes(os.path.expanduser(os.path.expandvars(x)), errors='surrogate_or_strict') for x in args if x is not None]
            else:
                args = [to_bytes(x, errors='surrogate_or_strict') for x in args if x is not None]

        prompt_re = None
        if prompt_regex:
            if isinstance(prompt_regex, str):
                prompt_regex = to_bytes(prompt_regex, errors='surrogateescape')
            try:
                prompt_re = re.compile(prompt_regex, re.MULTILINE)
            except re.error:
                self.fail_json(msg="invalid prompt regular expression given to run_command")

        rc = 0
        msg = None
        st_in = None

        env = os.environ.copy()
        # We can set this from both an attribute and per call
        env.update(self.run_command_environ_update or {})
        env.update(environ_update or {})
        if path_prefix:
            path = env.get('PATH', '')
            if path:
                env['PATH'] = "%s:%s" % (path_prefix, path)
            else:
                env['PATH'] = path_prefix

        # If using test-module.py and explode, the remote lib path will resemble:
        #   /tmp/test_module_scratch/debug_dir/ansible/module_utils/basic.py
        # If using ansible or ansible-playbook with a remote system:
        #   /tmp/ansible_vmweLQ/ansible_modlib.zip/ansible/module_utils/basic.py

        # Clean out python paths set by ansiballz
        if 'PYTHONPATH' in env:
            pypaths = [x for x in env['PYTHONPATH'].split(':')
                       if x and
                       not x.endswith('/ansible_modlib.zip') and
                       not x.endswith('/debug_dir')]
            if pypaths and any(pypaths):
                env['PYTHONPATH'] = ':'.join(pypaths)

        if data:
            st_in = subprocess.PIPE

        def preexec():
            if umask:
                os.umask(umask)

        kwargs = dict(
            executable=executable,
            shell=shell,
            close_fds=close_fds,
            stdin=st_in,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=preexec,
            env=env,
        )
        if pass_fds:
            kwargs["pass_fds"] = pass_fds

        # make sure we're in the right working directory
        if cwd:
            cwd = to_bytes(os.path.abspath(os.path.expanduser(cwd)), errors='surrogate_or_strict')
            if os.path.isdir(cwd):
                kwargs['cwd'] = cwd
            elif not ignore_invalid_cwd:
                self.fail_json(msg="Provided cwd is not a valid directory: %s" % cwd)

        try:
            if self._debug:
                self.log('Executing: ' + self._clean_args(args))
            cmd = subprocess.Popen(args, **kwargs)
            if before_communicate_callback:
                before_communicate_callback(cmd)

            stdout = b''
            stderr = b''

            # Mirror the CPython subprocess logic and preference for the selector to use.
            # poll/select have the advantage of not requiring any extra file
            # descriptor, contrarily to epoll/kqueue (also, they require a single
            # syscall).
            if hasattr(selectors, 'PollSelector'):
                selector = selectors.PollSelector()
            else:
                selector = selectors.SelectSelector()

            if data:
                if not binary_data:
                    data += '\n'
                if isinstance(data, str):
                    data = to_bytes(data)

            selector.register(cmd.stdout, selectors.EVENT_READ)
            selector.register(cmd.stderr, selectors.EVENT_READ)

            if os.name == 'posix':
                fcntl.fcntl(cmd.stdout.fileno(), fcntl.F_SETFL, fcntl.fcntl(cmd.stdout.fileno(), fcntl.F_GETFL) | os.O_NONBLOCK)
                fcntl.fcntl(cmd.stderr.fileno(), fcntl.F_SETFL, fcntl.fcntl(cmd.stderr.fileno(), fcntl.F_GETFL) | os.O_NONBLOCK)

            if data:
                cmd.stdin.write(data)
                cmd.stdin.close()

            while True:
                # A timeout of 1 is both a little short and a little long.
                # With None we could deadlock, with a lower value we would
                # waste cycles. As it is, this is a mild inconvenience if
                # we need to exit, and likely doesn't waste too many cycles
                events = selector.select(1)
                stdout_changed = False
                for key, event in events:
                    b_chunk = key.fileobj.read(32768)
                    if not b_chunk and b_chunk is not None:
                        selector.unregister(key.fileobj)
                    elif key.fileobj == cmd.stdout:
                        stdout += b_chunk
                        stdout_changed = True
                    elif key.fileobj == cmd.stderr:
                        stderr += b_chunk

                # if we're checking for prompts, do it now, but only if stdout
                # actually changed since the last loop
                if prompt_re and stdout_changed and prompt_re.search(stdout) and not data:
                    if encoding:
                        stdout = to_native(stdout, encoding=encoding, errors=errors)
                    return (257, stdout, "A prompt was encountered while running a command, but no input data was specified")

                # break out if no pipes are left to read or the pipes are completely read
                # and the process is terminated
                if (not events or not selector.get_map()) and cmd.poll() is not None:
                    break

                # No pipes are left to read but process is not yet terminated
                # Only then it is safe to wait for the process to be finished
                # NOTE: Actually cmd.poll() is always None here if no selectors are left
                elif not selector.get_map() and cmd.poll() is None:
                    cmd.wait()
                    # The process is terminated. Since no pipes to read from are
                    # left, there is no need to call select() again.
                    break

            cmd.stdout.close()
            cmd.stderr.close()
            selector.close()

            rc = cmd.returncode
        except OSError as ex:
            if handle_exceptions:
                self.fail_json(rc=ex.errno, stdout='', stderr='', msg="Error executing command.", cmd=self._clean_args(args), exception=ex)
            else:
                raise
        except Exception as ex:
            if handle_exceptions:
                self.fail_json(rc=257, stdout='', stderr='', msg="Error executing command.", cmd=self._clean_args(args), exception=ex)
            else:
                raise

        if rc != 0 and check_rc:
            msg = heuristic_log_sanitize(stderr.rstrip(), self.no_log_values)
            self.fail_json(cmd=self._clean_args(args), rc=rc, stdout=stdout, stderr=stderr, msg=msg)

        if encoding is not None:
            return (rc, to_native(stdout, encoding=encoding, errors=errors),
                    to_native(stderr, encoding=encoding, errors=errors))

        return (rc, stdout, stderr)

    def append_to_file(self, filename, str):
        filename = os.path.expandvars(os.path.expanduser(filename))
        with open(filename, 'a') as fh:
            fh.write(str)

    @staticmethod
    def bytes_to_human(size: int) -> str:
        return bytes_to_human(size)

    # for backwards compatibility
    pretty_bytes = bytes_to_human

    @staticmethod
    def human_to_bytes(number: str, isbits: bool = False) -> int:
        return human_to_bytes(number, isbits=isbits)

    #
    # Backwards compat
    #

    # In 2.0, moved from inside the module to the toplevel
    is_executable = is_executable

    @staticmethod
    def get_buffer_size(fd):
        try:
            # 1032 == FZ_GETPIPE_SZ
            buffer_size = fcntl.fcntl(fd, 1032)
        except Exception:
            try:
                # not as exact as above, but should be good enough for most platforms that fail the previous call
                buffer_size = select.PIPE_BUF
            except Exception:
                buffer_size = 9000  # use logical default JIC

        return buffer_size


def get_module_path():
    return os.path.dirname(os.path.realpath(__file__))


def __getattr__(importable_name):
    """Inject import-time deprecation warnings."""
    if importable_name == 'datetime':
        import datetime
        importable = datetime
    elif importable_name == 'signal':
        import signal
        importable = signal
    elif importable_name == 'types':
        import types
        importable = types
    elif importable_name == 'chain':
        from itertools import chain
        importable = chain
    elif importable_name == 'repeat':
        from itertools import repeat
        importable = repeat
    elif importable_name in {
        'PY2', 'PY3', 'b', 'binary_type', 'integer_types',
        'iteritems', 'string_types', 'text_type',
    }:
        import importlib
        importable = getattr(
            importlib.import_module('ansible.module_utils.six'),
            importable_name
        )
    elif importable_name == 'map':
        importable = map
    elif importable_name == 'shlex_quote':
        importable = shlex.quote
    else:
        raise AttributeError(
            f'cannot import name {importable_name !r} '
            f"from '{__name__}' ({__file__ !s})"
        )

    deprecate(
        msg=f"Importing '{importable_name}' from '{__name__}' is deprecated.",
        version="2.21",
    )
    return importable



from __future__ import annotations

import typing as t

import pytest

from ansible._internal._json import AnsibleVariableVisitor, EncryptedStringBehavior
from ansible.errors import AnsibleVariableTypeError
from ansible.parsing.vault import EncryptedString, AnsibleVaultError
from units.mock.vault_helper import VaultTestHelper


@pytest.mark.parametrize("behavior, decryptable, expected", (
    (EncryptedStringBehavior.PRESERVE, True, None),
    (EncryptedStringBehavior.PRESERVE, False, None),
    (EncryptedStringBehavior.DECRYPT, True, "plaintext"),
    (EncryptedStringBehavior.DECRYPT, False, AnsibleVaultError("no vault secrets")),
    (EncryptedStringBehavior.REDACT, True, "<redacted>"),
    (EncryptedStringBehavior.REDACT, False, "<redacted>"),
    (EncryptedStringBehavior.FAIL, True, AnsibleVariableTypeError("unsupported for variable storage")),
    (EncryptedStringBehavior.FAIL, False, AnsibleVariableTypeError("unsupported for variable storage")),
), ids=str)
def test_encrypted_string_behavior(
    behavior: EncryptedStringBehavior,
    decryptable: bool,
    expected: t.Any,
    _vault_secrets_context: None,
) -> None:
    if decryptable:
        value = VaultTestHelper.make_encrypted_string('plaintext')
    else:
        # valid ciphertext with intentionally unavailable secret
        value = EncryptedString(ciphertext=(
            '$ANSIBLE_VAULT;1.1;AES256\n'
            '333665623864636331356364306535613231613833616662656130613665336561316435393736366636663864396636326330626530643238653462333562350a396162623230643'
            '037396430383335386663363534353733386430643764303062633738613533336135653563313139373038333964316264633265376435370a326137363231646261303036356636'
            '37346430303361316436306130663461393832656134346639326365633830373361376236343961386164323538353962'
        ))

    avv = AnsibleVariableVisitor(encrypted_string_behavior=behavior)

    if isinstance(expected, Exception):
        with pytest.raises(type(expected), match=expected.args[0]):
            avv.visit(value)
    else:
        result = avv.visit(value)

        if expected is None:
            assert result is value
        else:
            assert result == expected



from __future__ import annotations

from ansible._internal._templating._jinja_common import Marker
from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    accept_args_markers = True

    def run(self, terms, variables, **kwargs):
        return [isinstance(term, Marker) for term in terms]



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# this is a windows documentation stub, actual code lives in the .ps1
# file of the same name

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r"""
---
module: win_wait_for
version_added: '2.4'
short_description: Waits for a condition before continuing
description:
- You can wait for a set amount of time C(timeout), this is the default if
  nothing is specified.
- Waiting for a port to become available is useful for when services are not
  immediately available after their init scripts return which is true of
  certain Java application servers.
- You can wait for a file to exist or not exist on the filesystem.
- This module can also be used to wait for a regex match string to be present
  in a file.
- You can wait for active connections to be closed before continuing on a
  local port.
options:
  connect_timeout:
    description:
    - The maximum number of seconds to wait for a connection to happen before
      closing and retrying.
    type: int
    default: 5
  delay:
    description:
    - The number of seconds to wait before starting to poll.
    type: int
  exclude_hosts:
    description:
    - The list of hosts or IPs to ignore when looking for active TCP
      connections when C(state=drained).
    type: list
  host:
    description:
    - A resolvable hostname or IP address to wait for.
    - If C(state=drained) then it will only check for connections on the IP
      specified, you can use '0.0.0.0' to use all host IPs.
    type: str
    default: '127.0.0.1'
  path:
    description:
    - The path to a file on the filesystem to check.
    - If C(state) is present or started then it will wait until the file
      exists.
    - If C(state) is absent then it will wait until the file does not exist.
    type: path
  port:
    description:
    - The port number to poll on C(host).
    type: int
  regex:
    description:
    - Can be used to match a string in a file.
    - If C(state) is present or started then it will wait until the regex
      matches.
    - If C(state) is absent then it will wait until the regex does not match.
    - Defaults to a multiline regex.
    type: str
    aliases: [ "search_regex", "regexp" ]
  sleep:
    description:
    - Number of seconds to sleep between checks.
    type: int
    default: 1
  state:
    description:
    - When checking a port, C(started) will ensure the port is open, C(stopped)
      will check that is it closed and C(drained) will check for active
      connections.
    - When checking for a file or a search string C(present) or C(started) will
      ensure that the file or string is present, C(absent) will check that the
      file or search string is absent or removed.
    type: str
    choices: [ absent, drained, present, started, stopped ]
    default: started
  timeout:
    description:
    - The maximum number of seconds to wait for.
    type: int
    default: 300
seealso:
- module: wait_for
- module: win_wait_for_process
author:
- Jordan Borean (@jborean93)
"""

EXAMPLES = r"""
- name: Wait 300 seconds for port 8000 to become open on the host, don't start checking for 10 seconds
  win_wait_for:
    port: 8000
    delay: 10

- name: Wait 150 seconds for port 8000 of any IP to close active connections
  win_wait_for:
    host: 0.0.0.0
    port: 8000
    state: drained
    timeout: 150

- name: Wait for port 8000 of any IP to close active connection, ignoring certain hosts
  win_wait_for:
    host: 0.0.0.0
    port: 8000
    state: drained
    exclude_hosts: ['10.2.1.2', '10.2.1.3']

- name: Wait for file C:\temp\log.txt to exist before continuing
  win_wait_for:
    path: C:\temp\log.txt

- name: Wait until process complete is in the file before continuing
  win_wait_for:
    path: C:\temp\log.txt
    regex: process complete

- name: Wait until file is removed
  win_wait_for:
    path: C:\temp\log.txt
    state: absent

- name: Wait until port 1234 is offline but try every 10 seconds
  win_wait_for:
    port: 1234
    state: absent
    sleep: 10
"""

RETURN = r"""
wait_attempts:
  description: The number of attempts to poll the file or port before module
    finishes.
  returned: always
  type: int
  sample: 1
elapsed:
  description: The elapsed seconds between the start of poll and the end of the
    module. This includes the delay if the option is set.
  returned: always
  type: float
  sample: 2.1406487
"""



from __future__ import annotations

import collections
import dataclasses
import math
import pathlib
import tempfile
import typing as t

import pytest

from ansible.parsing.vault import EncryptedString
from ansible.plugins.loader import test_loader
from ansible.plugins.test import AnsibleJinja2Test
from ansible.template import Templar, trust_as_template
from units.test_utils.controller.display import emits_warnings


@dataclasses.dataclass
class Extra:
    variables: dict[str, t.Any] | None = None
    args: list[t.Any] | None = None
    kwargs: dict[str, t.Any] | None = None
    func: t.Callable[[Extra], None] | None = None


class MakeLink:
    _tempdir: tempfile.TemporaryDirectory[str] | None = None

    def __call__(self, *args, **kwargs) -> str:
        self._tempdir = tempfile.TemporaryDirectory()

        symlink = pathlib.Path(self._tempdir.name) / 'a_symlink'
        symlink.symlink_to('something')

        return str(symlink)

    def __del__(self) -> None:
        if self._tempdir:
            self._tempdir.cleanup()

    def __repr__(self) -> str:
        return 'MakeLink'


TEST_DATA_SET: tuple[tuple[t.Any, str, bool, Extra | None], ...] = (
    # core
    (dict(failed=1), 'failed', True, None),
    (dict(failed=0), 'failed', False, None),
    (dict(), 'failed', False, None),
    (dict(failed=1), 'success', False, None),
    (dict(failed=0), 'success', True, None),
    (dict(), 'success', True, None),
    (dict(unreachable=1), 'reachable', False, None),
    (dict(unreachable=0), 'reachable', True, None),
    (dict(), 'reachable', True, None),
    (dict(unreachable=0), 'unreachable', False, None),
    (dict(unreachable=1), 'unreachable', True, None),
    (dict(), 'unreachable', False, None),
    (dict(timedout=dict(period=99)), 'timedout', True, None),
    # (dict(timedout=1), 'timedout', False, None),  # oops, bug
    (dict(timedout=0), 'timedout', False, None),
    (dict(), 'timedout', False, None),
    (dict(changed=1), 'changed', True, None),
    (dict(changed=0), 'changed', False, None),
    (dict(), 'changed', False, None),
    # (dict(results=[]), 'changed', True, None), # oops, bug
    (dict(results=[dict(changed=1)]), 'changed', True, None),
    (dict(results=[dict(changed=0)]), 'changed', False, None),
    (dict(), 'changed', False, None),
    (dict(skipped=1), 'skipped', True, None),
    (dict(skipped=0), 'skipped', False, None),
    (dict(), 'skipped', False, None),
    (dict(finished=1), 'finished', True, None),
    (dict(finished=0), 'finished', False, None),
    (dict(), 'finished', True, None),
    (dict(started=1), 'started', True, None),
    (dict(started=0), 'started', False, None),
    (dict(), 'started', True, None),
    ('"foo"', 'match', True, Extra(args=['"foo"'])),
    ('"foo"', 'match', False, Extra(args=['"bar"'])),
    ('"xxfooxx"', 'search', True, Extra(args=['"foo"'])),
    ('"xxfooxx"', 'search', False, Extra(args=['"bar"'])),
    ('"fooxx"', 'regex', True, Extra(args=['"FOO"'], kwargs=dict(ignorecase=True, multiline=True, match_type='"match"'))),
    ('"fooxx"', 'regex', False, Extra(args=['"BAR"'], kwargs=dict(ignorecase=True, multiline=True, match_type='"match"'))),
    ('1.1', 'version_compare', True, Extra(args=['1.1', '"eq"'])),
    ('1.1', 'version_compare', False, Extra(args=['1.0', '"eq"'])),
    ([0], 'any', False, None),
    ([1], 'any', True, None),
    ([0], 'all', False, None),
    ([1], 'all', True, None),
    (1, 'truthy', True, None),
    (0, 'truthy', False, None),
    (1, 'falsy', False, None),
    (0, 'falsy', True, None),
    ('foo', 'vault_encrypted', True, Extra(variables=dict(foo=EncryptedString(ciphertext='$ANSIBLE_VAULT;1.1;BLAH')))),
    ('foo', 'vault_encrypted', False, Extra(variables=dict(foo='not_encrypted'))),
    (repr(str(pathlib.Path(__file__).parent / "dummy_vault.txt")), 'vaulted_file', True, None),
    (repr(__file__), 'vaulted_file', False, None),
    ('q', 'defined', True, None),
    ('not_defined', 'defined', False, None),
    ('q', 'undefined', False, None),
    ('not_defined', 'undefined', True, None),
    # files
    ('"/"', 'directory', True, None),
    (repr(__file__), 'directory', False, None),
    (repr(__file__), 'file', True, None),
    ('"/"', 'file', False, None),
    ('make_link()', 'link', True, Extra(variables=dict(make_link=MakeLink()))),
    ('"/"', 'link', False, None),
    ('"/"', 'exists', True, None),
    ('"/does_not_exist"', 'exists', False, None),
    ('"/"', 'link_exists', True, None),
    ('"/does_not_exist"', 'link_exists', False, None),
    ('"/absolute"', 'abs', True, None),
    ('"relative"', 'abs', False, None),
    ('"/"', 'same_file', True, Extra(args=['"/"'])),
    (repr(__file__), 'same_file', False, Extra(args=['"/"'])),
    ('"/"', 'mount', True, None),
    ('"/not_a_mount_point"', 'mount', False, None),
    # mathstuff
    ([1], 'subset', True, Extra(args=[[1]])),
    ([0], 'subset', False, Extra(args=[[1]])),
    ([1], 'superset', True, Extra(args=[[1]])),
    ([0], 'superset', False, Extra(args=[[1]])),
    ([0], 'contains', True, Extra(args=[0])),
    ([1], 'contains', False, Extra(args=[0])),
    ('nan', 'nan', True, Extra(variables=dict(nan=math.nan))),
    ('"a string"', 'nan', False, None),
    # uri
    ('"https://ansible.com/"', 'uri', True, None),
    (1, 'uri', False, None),
    ('"https://ansible.com/"', 'url', True, None),
    (1, 'url', False, None),
    ('"urn:https://ansible.com/"', 'urn', True, None),
    (1, 'urn', False, None),
)


@pytest.mark.parametrize("value,test,expected,extra", TEST_DATA_SET, ids=str)
def test_truthy_inputs(value: object, test: str, expected: bool, extra: Extra | None) -> None:
    """Ensure test plugins return the expected bool result, not just a truthy/falsey value."""
    test_invocation = test

    if extra:
        test_args = extra.args or []
        test_args.extend(f'{k}={v}' for k, v in (extra.kwargs or {}).items())
        test_invocation += '(' + ', '.join(str(arg) for arg in test_args) + ')'

    expression = f'{value} is {test_invocation}'

    with emits_warnings(deprecation_pattern=[]):
        result = Templar(variables=extra.variables if extra else None).evaluate_expression(trust_as_template(expression))

    assert result is expected


def test_ensure_all_plugins_tested() -> None:
    """Ensure all plugins have at least one entry in the test data set, accounting for functions which have multiple names."""
    test_plugins: list[AnsibleJinja2Test] = [plugin for plugin in test_loader.all() if plugin.ansible_name.startswith('ansible.builtin.')]
    plugin_aliases: dict[t.Any, set[str]] = collections.defaultdict(set)

    for test_plugin in test_plugins:
        plugin_aliases[test_plugin.j2_function].add(test_plugin.ansible_name)

    missing_entries: list[str] = []

    for plugin_names in plugin_aliases.values():
        matching_tests = {_expected for _value, test, _expected, _extra in TEST_DATA_SET if f'ansible.builtin.{test}' in plugin_names}
        missing = {True, False} - matching_tests

        if missing:  # pragma: nocover
            missing_entries.append(f'{plugin_names}: {missing}')

    assert not missing_entries



#!/usr/bin/python
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = """
module: _not_deprecated
short_description: This module is not deprecated
description: Its name has a leading underscore, but it is not deprecated.
author:
  - Ansible Core Team
"""

EXAMPLES = """#"""
RETURN = """"""

from ansible.module_utils.basic import AnsibleModule


if __name__ == '__main__':
    module = AnsibleModule(argument_spec=dict())
    module.exit_json()



from __future__ import annotations

import http.server
import socketserver
import ssl

if __name__ == '__main__':
    Handler = http.server.SimpleHTTPRequestHandler
    context = ssl.SSLContext()
    context.load_cert_chain(certfile='./cert.pem', keyfile='./key.pem')
    httpd = socketserver.TCPServer(("", 4443), Handler)
    httpd.socket = context.wrap_socket(httpd.socket, server_side=True)

    httpd.serve_forever()



# -*- coding: utf-8 -*-
# Copyright: (c) 2018, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
from __future__ import annotations

import re
import shlex

from abc import abstractmethod
from secrets import choice
from string import ascii_lowercase
from gettext import dgettext

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_bytes
from ansible.plugins import AnsiblePlugin
from ansible.utils import display as _display


def _gen_id(length=32):
    """ return random string used to identify the current privilege escalation """
    return ''.join(choice(ascii_lowercase) for x in range(length))


class BecomeBase(AnsiblePlugin):

    name = None  # type: str | None

    # messages for detecting prompted password issues
    fail = tuple()  # type: tuple[str, ...]
    missing = tuple()  # type: tuple[str, ...]

    # many connection plugins cannot provide tty, set to True if your become
    # plugin requires a tty, i.e su
    require_tty = False

    # plugin allows for pipelining execution
    pipelining = True

    # prompt to match
    prompt = ''

    def __init__(self):
        super(BecomeBase, self).__init__()
        self._id = ''
        self.success = ''

    def get_option(self, option, hostvars=None, playcontext=None):
        """ Overrides the base get_option to provide a fallback to playcontext vars in case a 3rd party plugin did not
        implement the base become options required in Ansible. """
        # TODO: add deprecation warning for ValueError in devel that removes the playcontext fallback
        try:
            return super(BecomeBase, self).get_option(option, hostvars=hostvars)
        except KeyError:
            pc_fallback = ['become_user', 'become_pass', 'become_flags', 'become_exe']
            if option not in pc_fallback:
                raise

            return getattr(playcontext, option, None)

    def expect_prompt(self) -> bool:
        """This function assists connection plugins in determining if they need to wait for
        a prompt. Both a prompt and a password are required.
        """
        return bool(self.prompt and self.get_option('become_pass'))

    def _build_success_command(self, cmd, shell, noexe=False):
        if not all((cmd, shell, self.success)):
            return cmd

        try:
            cmd = shlex.quote('%s %s %s %s' % (shell.ECHO, self.success, shell.COMMAND_SEP, cmd))
        except AttributeError as ex:
            raise AnsibleError(f'The {shell._load_name!r} shell plugin does not support become. It is missing the {ex.name!r} attribute.')
        exe = getattr(shell, 'executable', None)
        if exe and not noexe:
            cmd = '%s -c %s' % (exe, cmd)
        return cmd

    @abstractmethod
    def build_become_command(self, cmd, shell):
        self._id = _gen_id()
        self.success = 'BECOME-SUCCESS-%s' % self._id

    def strip_become_prompt(self, data: bytes) -> bytes:
        """
        Strips the first found configured become prompt from `data`, trailing whitespace and anything that precedes the prompt, then returns the result.
        If no prompt is expected, or the prompt is not `str` or `bytes`, `data` will be returned as-is.
        """
        if not self.prompt or not isinstance(self.prompt, (str, bytes)) or not self.expect_prompt():
            return data

        return self._strip_through_prefix(self.prompt, data)

    def strip_become_success(self, data: bytes) -> bytes:
        """Strips the first found success marker from `data`, trailing whitespace and anything that precedes the success marker, then returns the result."""
        return self._strip_through_prefix(self.success, data)

    @staticmethod
    def _strip_through_prefix(match: str | bytes, data: bytes) -> bytes:
        """Strips the first occurrence of `match` from `data`, trailing whitespace and anything that precedes `match`, then returns the result."""
        return re.sub(br'^.*?' + re.escape(to_bytes(match)) + br'\s*', b'', data, count=1, flags=re.DOTALL)

    def check_success(self, b_output):
        b_success = to_bytes(self.success)
        return any(b_success in l.rstrip() for l in b_output.splitlines(True))

    def check_password_prompt(self, b_output):
        """ checks if the expected password prompt exists in b_output """
        if self.prompt:
            b_prompt = to_bytes(self.prompt).strip()
            return any(l.strip().startswith(b_prompt) for l in b_output.splitlines())
        return False

    def _check_password_error(self, b_out, msg):
        """ returns True/False if domain specific i18n version of msg is found in b_out """
        b_fail = to_bytes(dgettext(self.name, msg))
        return b_fail and b_fail in b_out

    def check_incorrect_password(self, b_output):
        for errstring in self.fail:
            if self._check_password_error(b_output, errstring):
                return True
        return False

    def check_missing_password(self, b_output):
        for errstring in self.missing:
            if self._check_password_error(b_output, errstring):
                return True
        return False



from __future__ import annotations

import threading

from ansible.module_utils._internal._concurrent import _daemon_threading


def test_daemon_thread_getattr() -> None:
    """Ensure that the threading module proxy delegates properly to the real module."""
    assert _daemon_threading.current_thread is threading.current_thread


def test_daemon_threading_thread_override() -> None:
    """Ensure that the proxy module's Thread attribute is different from the real module's."""
    assert _daemon_threading.Thread is not threading.Thread



from __future__ import annotations

import abc
import typing as t

from dataclasses import dataclass

# noinspection PyProtectedMember
from ansible.module_utils._internal._datatag import AnsibleTagHelper
# noinspection PyProtectedMember
from ansible._internal._templating._access import AnsibleAccessContext, NotifiableAccessContextBase

from units.module_utils.datatag.test_datatag import ExampleSingletonTag, ExampleTagWithContent

untagged_values = (123, 123.45, 'dude', ['dude'], dict(dude='mar'))


@dataclass(frozen=True)
class LoggedAccess:
    ctx: NotifiableAccessContextBase
    obj: object


class LoggingTagAccessNotifier(NotifiableAccessContextBase, metaclass=abc.ABCMeta):
    def __init__(self, access_list: list):
        self._access_list: list = access_list

    def _log(self, o: t.Any) -> t.Any:
        self._access_list.append(LoggedAccess(ctx=self, obj=o))


class ExampleTagWithContentAccessNotifier(LoggingTagAccessNotifier):
    _type_interest = frozenset([ExampleTagWithContent])

    def _notify(self, o: t.Any) -> t.Any:
        super()._log(o)  # get parent logging behavior
        return o


class ExampleSingletonTagAccessNotifier(LoggingTagAccessNotifier):
    _type_interest = frozenset([ExampleSingletonTag])

    def _notify(self, o: t.Any) -> t.Any:
        super()._log(o)  # get parent logging behavior
        return o


class ExampleMaskingSingletonTagAccessNotifier1(ExampleSingletonTagAccessNotifier):
    _mask = True


class ExampleMaskingSingletonTagAccessNotifier2(ExampleMaskingSingletonTagAccessNotifier1):
    ...


def test_ansibleaccesscontext_untagged():
    # accessing untagged objects should always succeed, be a no-op, and return the original value
    for v in untagged_values:
        AnsibleAccessContext.current().access(v)


def test_ansibleaccesscontext_notify():
    tagged_values = [AnsibleTagHelper.tag(v, [ExampleSingletonTag(), ExampleTagWithContent(content_str='replacement')]) for v in untagged_values]

    instance_access_list = []
    singleton_access_list = []

    with ExampleTagWithContentAccessNotifier(instance_access_list):
        with ExampleSingletonTagAccessNotifier(singleton_access_list):
            for tv in tagged_values:
                AnsibleAccessContext.current().access(tv)

    assert [v.obj for v in instance_access_list] == [v.obj for v in singleton_access_list] == tagged_values


def test_mixed_mask_unmask():
    """Ensure that only the innermost instance of each type of a masking access context is notified, while non-masking contexts are always notified."""
    value = ExampleSingletonTag().tag('blah')

    access_log = []

    with (ExampleSingletonTagAccessNotifier(access_log) as outer_nonmasked,
          ExampleMaskingSingletonTagAccessNotifier1(access_log),  # masked
          ExampleMaskingSingletonTagAccessNotifier2(access_log),
          ExampleMaskingSingletonTagAccessNotifier1(access_log) as inner_masked_1,
          ExampleMaskingSingletonTagAccessNotifier2(access_log) as inner_masked_2,
          ExampleSingletonTagAccessNotifier(access_log) as inner_nonmasked,
          ):
        AnsibleAccessContext.current().access(value)

    assert len(access_log) == 4
    assert access_log == [
        LoggedAccess(inner_nonmasked, value),
        LoggedAccess(inner_masked_2, value),
        LoggedAccess(inner_masked_1, value),
        LoggedAccess(outer_nonmasked, value)
    ]



# Copyright: (c) 2021, Ansible Project

from __future__ import annotations

from ansible.errors import AnsibleError
from ansible.module_utils.common.text.converters import to_native, to_bytes
from ansible._internal._templating._jinja_common import VaultExceptionMarker
from ansible._internal._datatag._tags import VaultedValue
from ansible.parsing.vault import is_encrypted, VaultSecret, VaultLib, VaultHelper
from ansible import template as _template
from ansible.utils.display import Display

display = Display()


def do_vault(data, secret, salt=None, vault_id='filter_default', wrap_object=False, vaultid=None):
    if not isinstance(secret, (str, bytes)):
        raise TypeError(f"Secret passed is required to be a string, instead we got {type(secret)}.")

    if not isinstance(data, (str, bytes)):
        raise TypeError(f"Can only vault strings, instead we got {type(data)}.")

    if vaultid is not None:
        display.deprecated(
            msg="Use of undocumented `vaultid`.",
            version="2.20",
            help_text="Use `vault_id` instead.",
        )

        if vault_id == 'filter_default':
            vault_id = vaultid
        else:
            display.warning("Ignoring vaultid as vault_id is already set.")

    vs = VaultSecret(to_bytes(secret))
    vl = VaultLib()
    try:
        vault = vl.encrypt(to_bytes(data), vs, vault_id, salt)
    except Exception as ex:
        raise AnsibleError("Unable to encrypt.") from ex

    if wrap_object:
        vault = VaultedValue(ciphertext=str(vault)).tag(secret)
    else:
        vault = to_native(vault)

    return vault


@_template.accept_args_markers
def do_unvault(vault, secret, vault_id='filter_default', vaultid=None):
    if isinstance(vault, VaultExceptionMarker):
        vault = vault._disarm()

    if (first_marker := _template.get_first_marker_arg((vault, secret, vault_id, vaultid), {})) is not None:
        return first_marker

    if not isinstance(secret, (str, bytes)):
        raise TypeError(f"Secret passed is required to be as string, instead we got {type(secret)}.")

    if not isinstance(vault, (str, bytes)):
        raise TypeError(f"Vault should be in the form of a string, instead we got {type(vault)}.")

    if vaultid is not None:
        display.deprecated(
            msg="Use of undocumented `vaultid`.",
            version="2.20",
            help_text="Use `vault_id` instead.",
        )

        if vault_id == 'filter_default':
            vault_id = vaultid
        else:
            display.warning("Ignoring vaultid as vault_id is already set.")

    vs = VaultSecret(to_bytes(secret))
    vl = VaultLib([(vault_id, vs)])

    if ciphertext := VaultHelper.get_ciphertext(vault, with_tags=True):
        vault = ciphertext

    if is_encrypted(vault):
        try:
            data = vl.decrypt(vault)
        except Exception as ex:
            raise AnsibleError("Unable to decrypt.") from ex
    else:
        data = vault

    return to_native(data)


class FilterModule(object):
    """ Ansible vault jinja2 filters """

    def filters(self):
        filters = {
            'vault': do_vault,
            'unvault': do_unvault,
        }

        return filters



"""Ansible-specific pylint plugin for checking deprecation calls."""

# (c) 2018, Matt Martz <matt@sivel.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import dataclasses
import datetime
import functools
import pathlib
import re

import astroid
import astroid.context
import astroid.typing

import pylint.lint
import pylint.checkers
import pylint.checkers.utils

import ansible.release

from ansible.module_utils._internal._deprecator import INDETERMINATE_DEPRECATOR, _path_as_collection_plugininfo
from ansible.module_utils.compat.version import StrictVersion
from ansible.utils.version import SemanticVersion


@dataclasses.dataclass(frozen=True, kw_only=True)
class DeprecationCallArgs:
    """Arguments passed to a deprecation function."""

    msg: object = None
    version: object = None
    date: object = None
    collection_name: object = None
    deprecator: object = None
    help_text: object = None  # only on Display.deprecated, warnings.deprecate and deprecate_value
    obj: object = None  # only on Display.deprecated and warnings.deprecate
    removed: object = None  # only on Display.deprecated
    value: object = None  # only on deprecate_value

    def all_args_dynamic(self) -> bool:
        """True if all args are dynamic or None, otherwise False."""
        return all(arg is None or isinstance(arg, astroid.NodeNG) for arg in dataclasses.asdict(self).values())


class AnsibleDeprecatedChecker(pylint.checkers.BaseChecker):
    """Checks for deprecated calls to ensure proper usage."""

    name = 'deprecated-calls'
    msgs = {
        'E9501': (
            "Deprecated version %r found in call to %r",
            "ansible-deprecated-version",
            None,
        ),
        'E9502': (
            "Found %r call without a version or date",
            "ansible-deprecated-no-version",
            None,
        ),
        'E9503': (
            "Invalid deprecated version %r found in call to %r",
            "ansible-invalid-deprecated-version",
            None,
        ),
        'E9504': (
            "Deprecated version %r found in call to %r",
            "collection-deprecated-version",
            None,
        ),
        'E9505': (
            "Invalid deprecated version %r found in call to %r",
            "collection-invalid-deprecated-version",
            None,
        ),
        'E9506': (
            "No collection_name or deprecator found in call to %r",
            "ansible-deprecated-no-collection-name",
            None,
        ),
        'E9507': (
            "Wrong collection_name %r found in call to %r",
            "wrong-collection-deprecated",
            None,
        ),
        'E9508': (
            "Expired date %r found in call to %r",
            "ansible-expired-deprecated-date",
            None,
        ),
        'E9509': (
            "Invalid date %r found in call to %r",
            "ansible-invalid-deprecated-date",
            None,
        ),
        'E9510': (
            "Both version and date found in call to %r",
            "ansible-deprecated-both-version-and-date",
            None,
        ),
        'E9511': (
            "Removal version %r must be a major release, not a minor or patch release, see https://semver.org/",
            "removal-version-must-be-major",
            None,
        ),
        'E9512': (
            "Passing date is not permitted in call to %r for ansible-core, use a version instead",
            "ansible-deprecated-date-not-permitted",
            None,
        ),
        'E9513': (
            "Unnecessary %r found in call to %r",
            "ansible-deprecated-unnecessary-collection-name",
            None,
        ),
        'E9514': (
            "Passing collection_name not permitted in call to %r for ansible-core, use deprecator instead",
            "ansible-deprecated-collection-name-not-permitted",
            None,
        ),
        'E9515': (
            "Both collection_name and deprecator found in call to %r",
            "ansible-deprecated-both-collection-name-and-deprecator",
            None,
        ),
    }

    options = (
        (
            'collection-name',
            dict(
                default=None,
                type='string',
                metavar='<name>',
                help="The name of the collection to check.",
            ),
        ),
        (
            'collection-version',
            dict(
                default=None,
                type='string',
                metavar='<version>',
                help="The version of the collection to check.",
            ),
        ),
        (
            'collection-path',
            dict(
                default=None,
                type='string',
                metavar='<path>',
                help="The path of the collection to check.",
            ),
        ),
    )

    ANSIBLE_VERSION = StrictVersion(re.match('[0-9.]*[0-9]', ansible.release.__version__)[0])
    """The current ansible-core X.Y.Z version."""

    DEPRECATION_MODULE_FUNCTIONS: dict[tuple[str, str], tuple[str, ...]] = {
        ('ansible.module_utils.common.warnings', 'deprecate'): ('msg', 'version', 'date', 'collection_name'),
        ('ansible.module_utils.datatag', 'deprecate_value'): ('value', 'msg'),
        ('ansible.module_utils.basic', 'AnsibleModule.deprecate'): ('msg', 'version', 'date', 'collection_name'),
        ('ansible.utils.display', 'Display.deprecated'): ('msg', 'version', 'removed', 'date', 'collection_name'),
    }
    """Mapping of deprecation module+function and their positional arguments."""

    DEPRECATION_MODULES = frozenset(key[0] for key in DEPRECATION_MODULE_FUNCTIONS)
    """Modules which contain deprecation functions."""

    DEPRECATION_FUNCTIONS = {'.'.join(key): value for key, value in DEPRECATION_MODULE_FUNCTIONS.items()}
    """Mapping of deprecation functions and their positional arguments."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.module_cache: dict[str, astroid.Module] = {}

    @functools.cached_property
    def collection_name(self) -> str | None:
        """Return the collection name, or None if ansible-core is being tested."""
        return self.linter.config.collection_name or None

    @functools.cached_property
    def collection_path(self) -> pathlib.Path:
        """Return the collection path. Not valid when ansible-core is being tested."""
        return pathlib.Path(self.linter.config.collection_path)

    @functools.cached_property
    def collection_version(self) -> SemanticVersion | None:
        """Return the collection version, or None if ansible-core is being tested."""
        if not self.linter.config.collection_version:
            return None

        sem_ver = SemanticVersion(self.linter.config.collection_version)
        sem_ver.prerelease = ()  # ignore pre-release for version comparison to catch issues before the final release is cut

        return sem_ver

    @functools.cached_property
    def is_ansible_core(self) -> bool:
        """True if ansible-core is being tested."""
        return not self.collection_name

    @functools.cached_property
    def today_utc(self) -> datetime.date:
        """Today's date in UTC."""
        return datetime.datetime.now(tz=datetime.timezone.utc).date()

    def is_deprecator_required(self) -> bool | None:
        """Determine is a `collection_name` or `deprecator` is required (True), unnecessary (False) or optional (None)."""
        if self.is_ansible_core:
            return False  # in ansible-core, never provide the deprecator -- if it really is needed, disable the sanity test inline for that line of code

        plugin_info = _path_as_collection_plugininfo(self.linter.current_file)

        if plugin_info is INDETERMINATE_DEPRECATOR:
            return True  # deprecator cannot be detected, caller must provide deprecator

        # deprecation: description='deprecate collection_name/deprecator now that detection is widely available' core_version='2.23'
        # When this deprecation triggers, change the return type here to False.
        # At that point, callers should be able to omit the collection_name/deprecator in all but a few cases (inline ignores can be used for those cases)
        return None

    @pylint.checkers.utils.only_required_for_messages(*(msgs.keys()))
    def visit_call(self, node: astroid.Call) -> None:
        """Visit a call node."""
        if inferred := self.infer(node.func):
            name = self.get_fully_qualified_name(inferred)

            if args := self.DEPRECATION_FUNCTIONS.get(name):
                self.check_call(node, name, args)

    def infer(self, node: astroid.NodeNG) -> astroid.NodeNG | None:
        """Return the inferred node from the given node, or `None` if it cannot be unambiguously inferred."""
        names: list[str] = []
        target: astroid.NodeNG | None = node
        inferred: astroid.typing.InferenceResult | None = None

        while target:
            if inferred := astroid.util.safe_infer(target):
                break

            if isinstance(target, astroid.Call):
                inferred = self.infer(target.func)
                break

            if isinstance(target, astroid.FunctionDef):
                inferred = target
                break

            if isinstance(target, astroid.Name):
                target = self.infer_name(target)
            elif isinstance(target, astroid.AssignName) and isinstance(target.parent, astroid.Assign):
                target = target.parent.value
            elif isinstance(target, astroid.Attribute):
                names.append(target.attrname)
                target = target.expr
            else:
                break

        for name in reversed(names):
            if isinstance(inferred, astroid.Instance):
                try:
                    attr = next(iter(inferred.getattr(name)), None)
                except astroid.AttributeInferenceError:
                    break

                if isinstance(attr, astroid.AssignAttr):
                    inferred = self.get_ansible_module(attr)
                    continue

                if isinstance(attr, astroid.FunctionDef):
                    inferred = attr
                    continue

            if not isinstance(inferred, (astroid.Module, astroid.ClassDef)):
                inferred = None
                break

            try:
                inferred = inferred[name]
            except KeyError:
                inferred = None
            else:
                inferred = self.infer(inferred)

        if isinstance(inferred, astroid.FunctionDef) and isinstance(inferred.parent, astroid.ClassDef):
            inferred = astroid.BoundMethod(inferred, inferred.parent)

        return inferred

    def infer_name(self, node: astroid.Name) -> astroid.NodeNG | None:
        """Infer the node referenced by the given name, or `None` if it cannot be unambiguously inferred."""
        scope = node.scope()
        inferred: astroid.NodeNG | None = None
        name = node.name

        while scope:
            try:
                assignment = scope[name]
            except KeyError:
                scope = scope.parent.scope() if scope.parent else None
                continue

            if isinstance(assignment, astroid.AssignName) and isinstance(assignment.parent, astroid.Assign):
                inferred = assignment.parent.value
            elif (
                isinstance(scope, astroid.FunctionDef)
                and isinstance(assignment, astroid.AssignName)
                and isinstance(assignment.parent, astroid.Arguments)
                and assignment.parent.annotations
            ):
                idx, _node = assignment.parent.find_argname(name)

                if idx is not None:
                    try:
                        annotation = assignment.parent.annotations[idx]
                    except IndexError:
                        pass
                    else:
                        if isinstance(annotation, astroid.Name):
                            name = annotation.name
                            continue
            elif isinstance(assignment, astroid.ClassDef):
                inferred = assignment
            elif isinstance(assignment, astroid.ImportFrom):
                if module := self.get_module(assignment):
                    name = assignment.real_name(name)
                    scope = module.scope()
                    continue

            break

        return inferred

    def get_module(self, node: astroid.ImportFrom) -> astroid.Module | None:
        """Import the requested module if possible and cache the result."""
        module_name = pylint.checkers.utils.get_import_name(node, node.modname)

        if module_name not in self.DEPRECATION_MODULES:
            return None  # avoid unnecessary import overhead

        if module := self.module_cache.get(module_name):
            return module

        module = node.do_import_module()

        if module.name != module_name:
            raise RuntimeError(f'Attempted to import {module_name!r} but found {module.name!r} instead.')

        self.module_cache[module_name] = module

        return module

    @staticmethod
    def get_fully_qualified_name(node: astroid.NodeNG) -> str | None:
        """Return the fully qualified name of the given inferred node."""
        parent = node.parent
        parts: tuple[str, ...] | None

        if isinstance(node, astroid.FunctionDef) and isinstance(parent, astroid.Module):
            parts = (parent.name, node.name)
        elif isinstance(node, astroid.BoundMethod) and isinstance(parent, astroid.ClassDef) and isinstance(parent.parent, astroid.Module):
            parts = (parent.parent.name, parent.name, node.name)
        else:
            parts = None

        return '.'.join(parts) if parts else None

    def check_call(self, node: astroid.Call, name: str, args: tuple[str, ...]) -> None:
        """Check the given deprecation call node for valid arguments."""
        call_args = self.get_deprecation_call_args(node, args)

        self.check_collection_name(node, name, call_args)

        if not call_args.version and not call_args.date:
            self.add_message('ansible-deprecated-no-version', node=node, args=(name,))
            return

        if call_args.date and self.is_ansible_core:
            self.add_message('ansible-deprecated-date-not-permitted', node=node, args=(name,))
            return

        if call_args.all_args_dynamic():
            # assume collection maintainers know what they're doing if all args are dynamic
            return

        if call_args.version and call_args.date:
            self.add_message('ansible-deprecated-both-version-and-date', node=node, args=(name,))
            return

        if call_args.date:
            self.check_date(node, name, call_args)

        if call_args.version:
            self.check_version(node, name, call_args)

    @staticmethod
    def get_deprecation_call_args(node: astroid.Call, args: tuple[str, ...]) -> DeprecationCallArgs:
        """Get the deprecation call arguments from the given node."""
        fields: dict[str, object] = {}

        for idx, arg in enumerate(node.args):
            field = args[idx]
            fields[field] = arg

        for keyword in node.keywords:
            if keyword.arg is not None:
                fields[keyword.arg] = keyword.value

        for key, value in fields.items():
            if isinstance(value, astroid.Const):
                fields[key] = value.value

        return DeprecationCallArgs(**fields)

    def check_collection_name(self, node: astroid.Call, name: str, args: DeprecationCallArgs) -> None:
        """Check the collection name provided to the given call node."""
        deprecator_requirement = self.is_deprecator_required()

        if self.is_ansible_core and args.collection_name:
            self.add_message('ansible-deprecated-collection-name-not-permitted', node=node, args=(name,))
            return

        if args.collection_name and args.deprecator:
            self.add_message('ansible-deprecated-both-collection-name-and-deprecator', node=node, args=(name,))

        if deprecator_requirement is True:
            if not args.collection_name and not args.deprecator:
                self.add_message('ansible-deprecated-no-collection-name', node=node, args=(name,))
                return
        elif deprecator_requirement is False:
            if args.collection_name:
                self.add_message('ansible-deprecated-unnecessary-collection-name', node=node, args=('collection_name', name,))
                return

            if args.deprecator:
                self.add_message('ansible-deprecated-unnecessary-collection-name', node=node, args=('deprecator', name,))
                return
        else:
            # collection_name may be needed for backward compat with 2.18 and earlier, since it is only detected in 2.19 and later

            if args.deprecator:
                # Unlike collection_name, which is needed for backward compat, deprecator is generally not needed by collections.
                # For the very rare cases where this is needed by collections, an inline pylint ignore can be used to silence it.
                self.add_message('ansible-deprecated-unnecessary-collection-name', node=node, args=('deprecator', name,))
                return

        if args.all_args_dynamic():
            # assume collection maintainers know what they're doing if all args are dynamic
            return

        expected_collection_name = 'ansible.builtin' if self.is_ansible_core else self.collection_name

        if args.collection_name and args.collection_name != expected_collection_name:
            self.add_message('wrong-collection-deprecated', node=node, args=(args.collection_name, name))

    def check_version(self, node: astroid.Call, name: str, args: DeprecationCallArgs) -> None:
        """Check the version provided to the given call node."""
        if self.collection_name:
            self.check_collection_version(node, name, args)
        else:
            self.check_core_version(node, name, args)

    def check_core_version(self, node: astroid.Call, name: str, args: DeprecationCallArgs) -> None:
        """Check the core version provided to the given call node."""
        try:
            if not isinstance(args.version, str) or not args.version:
                raise ValueError()

            strict_version = StrictVersion(args.version)
        except ValueError:
            self.add_message('ansible-invalid-deprecated-version', node=node, args=(args.version, name))
            return

        if self.ANSIBLE_VERSION >= strict_version:
            self.add_message('ansible-deprecated-version', node=node, args=(args.version, name))

    def check_collection_version(self, node: astroid.Call, name: str, args: DeprecationCallArgs) -> None:
        """Check the collection version provided to the given call node."""
        try:
            if not isinstance(args.version, str) or not args.version:
                raise ValueError()

            semantic_version = SemanticVersion(args.version)
        except ValueError:
            self.add_message('collection-invalid-deprecated-version', node=node, args=(args.version, name))
            return

        if self.collection_version >= semantic_version:
            self.add_message('collection-deprecated-version', node=node, args=(args.version, name))

        if semantic_version.major != 0 and (semantic_version.minor != 0 or semantic_version.patch != 0):
            self.add_message('removal-version-must-be-major', node=node, args=(args.version,))

    def check_date(self, node: astroid.Call, name: str, args: DeprecationCallArgs) -> None:
        """Check the date provided to the given call node."""
        try:
            date_parsed = self.parse_isodate(args.date)
        except (ValueError, TypeError):
            self.add_message('ansible-invalid-deprecated-date', node=node, args=(args.date, name))
        else:
            if date_parsed < self.today_utc:
                self.add_message('ansible-expired-deprecated-date', node=node, args=(args.date, name))

    @staticmethod
    def parse_isodate(value: object) -> datetime.date:
        """Parse an ISO 8601 date string."""
        if isinstance(value, str):
            return datetime.date.fromisoformat(value)

        raise TypeError(type(value))

    def get_ansible_module(self, node: astroid.AssignAttr) -> astroid.Instance | None:
        """Infer an AnsibleModule instance node from the given assignment."""
        if isinstance(node.parent, astroid.Assign) and isinstance(node.parent.type_annotation, astroid.Name):
            inferred = self.infer_name(node.parent.type_annotation)
        elif isinstance(node.parent, astroid.Assign) and isinstance(node.parent.parent, astroid.FunctionDef) and isinstance(node.parent.value, astroid.Name):
            inferred = self.infer_name(node.parent.value)
        elif isinstance(node.parent, astroid.AnnAssign) and isinstance(node.parent.annotation, astroid.Name):
            inferred = self.infer_name(node.parent.annotation)
        else:
            inferred = None

        if isinstance(inferred, astroid.ClassDef) and inferred.name == 'AnsibleModule':
            return inferred.instantiate_class()

        return None

    def register(self) -> None:
        """Register this plugin."""
        self.linter.register_checker(self)


def register(linter: pylint.lint.PyLinter) -> None:
    """Required method to auto-register this checker."""
    AnsibleDeprecatedChecker(linter).register()



from __future__ import annotations

import abc
import collections.abc as c
import enum
import inspect
import itertools
import typing as t

from jinja2 import UndefinedError, StrictUndefined, TemplateRuntimeError
from jinja2.utils import missing

from ...module_utils._internal import _messages
from ansible.constants import config
from ansible.errors import AnsibleUndefinedVariable, AnsibleTypeError
from ansible._internal._errors._handler import ErrorHandler
from ansible.module_utils._internal._datatag import Tripwire, _untaggable_types

from ._access import NotifiableAccessContextBase
from ._jinja_patches import _patch_jinja
from ._utils import TemplateContext
from .._errors import _captured
from ...module_utils.datatag import native_type_name

_patch_jinja()  # apply Jinja2 patches before types are declared that are dependent on the changes


class _SandboxMode(enum.Enum):
    DEFAULT = enum.auto()
    ALLOW_UNSAFE_ATTRIBUTES = enum.auto()


class _TemplateConfig:
    allow_embedded_templates: bool = config.get_config_value("ALLOW_EMBEDDED_TEMPLATES")
    allow_broken_conditionals: bool = config.get_config_value('ALLOW_BROKEN_CONDITIONALS')
    jinja_extensions: list[str] = config.get_config_value('DEFAULT_JINJA2_EXTENSIONS')
    sandbox_mode: _SandboxMode = _SandboxMode.__members__[config.get_config_value('_TEMPLAR_SANDBOX_MODE').upper()]

    unknown_type_encountered_handler = ErrorHandler.from_config('_TEMPLAR_UNKNOWN_TYPE_ENCOUNTERED')
    unknown_type_conversion_handler = ErrorHandler.from_config('_TEMPLAR_UNKNOWN_TYPE_CONVERSION')
    untrusted_template_handler = ErrorHandler.from_config('_TEMPLAR_UNTRUSTED_TEMPLATE_BEHAVIOR')


class MarkerError(UndefinedError):
    """
    An Ansible specific subclass of Jinja's UndefinedError, used to preserve and later restore the original Marker instance that raised the error.
    This error is only raised by Marker and should never escape the templating system.
    """

    def __init__(self, message: str, source: Marker) -> None:
        super().__init__(message)

        self.source = source


class Marker(StrictUndefined, Tripwire):
    """
    Extends Jinja's `StrictUndefined`, allowing any kind of error occurring during recursive templating operations to be captured and deferred.
    Direct or managed access to most `Marker` attributes will raise a `MarkerError`, which usually ends the current innermost templating
    operation and converts the `MarkerError` back to the origin Marker instance (subject to the `MarkerBehavior` in effect at the time).
    """

    __slots__ = ('_marker_template_source',)

    _concrete_subclasses: t.ClassVar[set[type[Marker]]] = set()

    def __init__(
        self,
        hint: t.Optional[str] = None,
        obj: t.Any = missing,
        name: t.Optional[str] = None,
        exc: t.Type[TemplateRuntimeError] = UndefinedError,  # Ansible doesn't set this argument or consume the attribute it is stored under.
        *args,
        _no_template_source=False,
        **kwargs,
    ) -> None:
        if not hint and name and obj is not missing:
            hint = f"object of type {native_type_name(obj)!r} has no attribute {name!r}"

        kwargs.update(
            hint=hint,
            obj=obj,
            name=name,
            exc=exc,
        )

        super().__init__(*args, **kwargs)

        if _no_template_source:
            self._marker_template_source = None
        else:
            self._marker_template_source = TemplateContext.current().template_value

    def _as_exception(self) -> Exception:
        """Return the exception instance to raise in a top-level templating context."""
        return AnsibleUndefinedVariable(self._undefined_message, obj=self._marker_template_source)

    def _as_message(self) -> str:
        """Return the error message to show when this marker must be represented as a string, such as for substitutions or warnings."""
        return self._undefined_message

    def _fail_with_undefined_error(self, *args: t.Any, **kwargs: t.Any) -> t.NoReturn:
        """Ansible-specific replacement for Jinja's _fail_with_undefined_error tripwire on dunder methods."""
        self.trip()

    def trip(self) -> t.NoReturn:
        """Raise an internal exception which can be converted back to this instance."""
        raise MarkerError(self._undefined_message, self)

    def __setattr__(self, name: str, value: t.Any) -> None:
        """
        Any attempt to set an unknown attribute on a `Marker` should invoke the trip method to propagate the original context.
        This does not protect against mutation of known attributes, but the implementation is fairly simple.
        """
        try:
            super().__setattr__(name, value)
        except AttributeError:
            pass
        else:
            return

        self.trip()

    def __getattr__(self, name: str) -> t.Any:
        """Raises AttributeError for dunder-looking accesses, self-propagates otherwise."""
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError(name)

        return self

    def __getitem__(self, key):
        """Self-propagates on all item accesses."""
        return self

    @classmethod
    def __init_subclass__(cls, **kwargs) -> None:
        if not inspect.isabstract(cls):
            _untaggable_types.add(cls)
            cls._concrete_subclasses.add(cls)

    @classmethod
    def _init_class(cls):
        _untaggable_types.add(cls)

        # These are the methods StrictUndefined already intercepts.
        jinja_method_names = (
            '__add__',
            '__bool__',
            '__call__',
            '__complex__',
            '__contains__',
            '__div__',
            '__eq__',
            '__float__',
            '__floordiv__',
            '__ge__',
            # '__getitem__',  # using a custom implementation that propagates self instead
            '__gt__',
            '__hash__',
            '__int__',
            '__iter__',
            '__le__',
            '__len__',
            '__lt__',
            '__mod__',
            '__mul__',
            '__ne__',
            '__neg__',
            '__pos__',
            '__pow__',
            '__radd__',
            '__rdiv__',
            '__rfloordiv__',
            '__rmod__',
            '__rmul__',
            '__rpow__',
            '__rsub__',
            '__rtruediv__',
            '__str__',
            '__sub__',
            '__truediv__',
        )

        # These additional methods should be intercepted, even though they are not intercepted by StrictUndefined.
        additional_method_names = (
            '__aiter__',
            '__delattr__',
            '__format__',
            '__repr__',
            '__setitem__',
        )

        for name in jinja_method_names + additional_method_names:
            setattr(cls, name, cls._fail_with_undefined_error)


Marker._init_class()


class TruncationMarker(Marker):
    """
    An `Marker` value was previously encountered and reported.
    A subsequent `Marker` value (this instance) indicates the template may have been truncated as a result.
    It will only be visible if the previous `Marker` was ignored/replaced instead of being tripped, which would raise an exception.
    """

    __slots__ = ()

    def __init__(self) -> None:
        super().__init__(hint='template potentially truncated')


class UndefinedMarker(Marker):
    """A `Marker` value that represents an undefined value encountered during templating."""

    __slots__ = ()


class ExceptionMarker(Marker, metaclass=abc.ABCMeta):
    """Base `Marker` class that represents exceptions encountered and deferred during templating."""

    __slots__ = ()

    @abc.abstractmethod
    def _as_exception(self) -> Exception:
        pass

    def _as_message(self) -> str:
        return str(self._as_exception())

    def trip(self) -> t.NoReturn:
        """Raise an internal exception which can be converted back to this instance while maintaining the cause for callers that follow them."""
        raise MarkerError(self._undefined_message, self) from self._as_exception()


class CapturedExceptionMarker(ExceptionMarker):
    """A `Marker` value that represents an exception raised during templating."""

    __slots__ = ('_marker_captured_exception',)

    def __init__(self, exception: Exception) -> None:
        super().__init__(hint=f'A captured exception marker was tripped: {exception}')

        self._marker_captured_exception = exception

    def _as_exception(self) -> Exception:
        return self._marker_captured_exception


class UndecryptableVaultError(_captured.AnsibleCapturedError):
    """Template-external error raised by VaultExceptionMarker when an undecryptable variable is accessed."""

    context = 'vault'
    _default_message = "Attempt to use undecryptable variable."


class VaultExceptionMarker(ExceptionMarker):
    """A `Marker` value that represents an error accessing a vaulted value during templating."""

    __slots__ = ('_marker_undecryptable_ciphertext', '_marker_event')

    def __init__(self, ciphertext: str, event: _messages.Event) -> None:
        super().__init__(hint='A vault exception marker was tripped.')

        self._marker_undecryptable_ciphertext = ciphertext
        self._marker_event = event

    def _as_exception(self) -> Exception:
        return UndecryptableVaultError(
            obj=self._marker_undecryptable_ciphertext,
            event=self._marker_event,
        )

    def _disarm(self) -> str:
        return self._marker_undecryptable_ciphertext


def get_first_marker_arg(args: c.Sequence, kwargs: dict[str, t.Any]) -> Marker | None:
    """Utility method to inspect plugin args and return the first `Marker` encountered, otherwise `None`."""
    # CAUTION: This function is exposed in public API as ansible.template.get_first_marker_arg.
    return next(iter_marker_args(args, kwargs), None)


def iter_marker_args(args: c.Sequence, kwargs: dict[str, t.Any]) -> t.Generator[Marker]:
    """Utility method to iterate plugin args and yield any `Marker` encountered."""
    for arg in itertools.chain(args, kwargs.values()):
        if isinstance(arg, Marker):
            yield arg


class JinjaCallContext(NotifiableAccessContextBase):
    """
    An audit context that wraps all Jinja (template/filter/test/lookup/method/function) calls.
    While active, calls `trip()` on managed access of `Marker` objects unless the callee declares an understanding of markers.
    """

    _mask = True

    def __init__(self, accept_lazy_markers: bool) -> None:
        self._type_interest = frozenset() if accept_lazy_markers else frozenset(Marker._concrete_subclasses)

    def _notify(self, o: Marker) -> t.NoReturn:
        o.trip()


def validate_arg_type(name: str, value: t.Any, allowed_type_or_types: type | tuple[type, ...], /) -> None:
    """Validate the type of the given argument while preserving context for Marker values."""
    # DTFIX-FUTURE: find a home for this as a general-purpose utliity method and expose it after some API review
    if isinstance(value, allowed_type_or_types):
        return

    if isinstance(allowed_type_or_types, type):
        arg_type_description = repr(native_type_name(allowed_type_or_types))
    else:
        arg_type_description = ' or '.join(repr(native_type_name(item)) for item in allowed_type_or_types)

    if isinstance(value, Marker):
        try:
            value.trip()
        except Exception as ex:
            raise AnsibleTypeError(f"The {name!r} argument must be of type {arg_type_description}.", obj=value) from ex

    raise AnsibleTypeError(f"The {name!r} argument must be of type {arg_type_description}, not {native_type_name(value)!r}.", obj=value)



# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

import pathlib

import pytest

from ansible.module_utils.facts.hardware import darwin
from ansible.module_utils.facts.sysctl import get_sysctl


class TestDarwinHardwareFacts:
    def _get_mock_sysctl_data(self, filename="sysctl_darwin_silicon.txt"):
        fixture_file = pathlib.Path(__file__).parent / "fixtures" / filename
        return fixture_file.read_text()

    @pytest.fixture()
    def mocked_module(self, mocker, request):
        request.cls.module = mocker.MagicMock()
        request.cls.module.get_bin_path.return_value = "/usr/sbin/sysctl"
        yield request.cls.module

    def test_get_mac_facts(self, mocked_module):
        mocked_module.run_command.return_value = (0, self._get_mock_sysctl_data(), "")
        darwin_hardware = darwin.DarwinHardware(mocked_module)
        darwin_hardware.sysctl = get_sysctl(
            mocked_module, ["hw", "machdep", "kern", "hw.model"]
        )

        mac_facts = darwin_hardware.get_mac_facts()
        expected_mac_facts = {
            "model": "MacBookPro18,1",
            "product_name": "MacBookPro18,1",
            "osversion": "23E224",
            "osrevision": "199506",
        }
        assert mac_facts == expected_mac_facts

    def test_get_cpu_facts(self, mocked_module):
        mocked_module.run_command.return_value = (0, self._get_mock_sysctl_data(), "")
        darwin_hardware = darwin.DarwinHardware(mocked_module)
        darwin_hardware.sysctl = get_sysctl(
            mocked_module, ["hw", "machdep", "kern", "hw.model"]
        )

        cpu_facts = darwin_hardware.get_cpu_facts()
        expected_cpu_facts = {
            "processor": "Apple M1 Pro",
            "processor_cores": "10",
            "processor_vcpus": "10",
        }
        assert cpu_facts == expected_cpu_facts

    @pytest.mark.parametrize(
        ("vm_stat_file", "sysctl_file", "expected_memory_facts"),
        [
            pytest.param(
                "vm_stat_darwin_intel.txt",
                "sysctl_darwin_intel.txt",
                {'memtotal_mb': 2048, 'memfree_mb': 178},
                id="intel",
            ),
            pytest.param(
                "vm_stat_darwin_silicon.txt",
                "sysctl_darwin_silicon.txt",
                {"memtotal_mb": 32768, "memfree_mb": 7660},
                id="silicon",
            ),
        ],
    )
    def test_get_memory_facts(self, mocked_module, vm_stat_file, sysctl_file, expected_memory_facts):
        fixtures = pathlib.Path(__file__).parent / "fixtures"
        mocked_module.get_bin_path.side_effect = [
            "/usr/sbin/sysctl",
            "/usr/bin/vm_stat",
        ]
        mocked_vm_stat = (fixtures / vm_stat_file).read_text()
        mocked_module.run_command.side_effect = [
            (0, self._get_mock_sysctl_data(filename=sysctl_file), ""),
            (0, mocked_vm_stat, ""),
        ]
        darwin_hardware = darwin.DarwinHardware(mocked_module)
        darwin_hardware.sysctl = get_sysctl(
            mocked_module, ["hw", "machdep", "kern", "hw.model"]
        )

        memory_facts = darwin_hardware.get_memory_facts()
        assert memory_facts == expected_memory_facts

    def test_get_uptime_facts(self, mocked_module):
        darwin_hardware = darwin.DarwinHardware(mocked_module)
        mocked_module.run_command.return_value = (
            0,
            b"\xc0\xa0\x05f\x00\x00\x00\x00\xac-\x05\x00\x00\x00\x00\x00",
            "",
        )
        uptime_facts = darwin_hardware.get_uptime_facts()
        assert "uptime_seconds" in uptime_facts



from __future__ import annotations

import pathlib
import tempfile

import pytest

from ansible.errors import AnsibleJSONParserError
from ansible._internal._errors._error_utils import format_exception_message
from ansible._internal._datatag._tags import Origin
from ansible.parsing.utils.yaml import from_yaml


def test_json_parser_error() -> None:
    # This is basically a copy of test_yaml_parser_error in test/units/parsing/yaml/test_errors.py.
    # Most of the coverage for parsing.utils.yaml is achieved while testing parsing.yaml.errors.

    content = 'x'
    expected_message = 'JSON parsing failed: Expecting value: line 1 column 1 (char 0)'
    line = 1
    col = 1
    expect_help_text = False

    with tempfile.TemporaryDirectory() as tempdir:
        source_path = pathlib.Path(tempdir) / 'source.yml'
        source_path.write_text(str(content))

        with pytest.raises(AnsibleJSONParserError) as error:
            from_yaml(content, file_name=str(source_path), json_only=True)

    assert error.value.message == expected_message
    assert error.value._original_message == expected_message
    assert format_exception_message(error.value) == expected_message
    assert str(error.value) == expected_message

    assert error.value.obj == Origin(path=str(source_path), line_num=line, col_num=col)

    if expect_help_text:
        assert error.value._help_text is not None  # DTFIX-FUTURE: check the content later once it's less volatile
    else:
        assert error.value._help_text is None



from __future__ import annotations

from ansible.plugins.lookup import LookupBase


class LookupModule(LookupBase):
    def run(self, terms, variables=None, **kwargs) -> list:
        raise NotImplementedError()
